# Copyright 2009 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['ConfigDialog']

import re

import gobject
import gtk
from gtk import gdk
import gtksourceview2
import pango

from .SimpleGladeApp import SimpleGladeApp
from . import tags
from .tags import DEFAULT, FG, BG, COLOR, ISSET
from .common import beep, get_text
from .file_dialogs import open_dialog

# Allow future translations
_ = lambda s: s

class ConfigDialog(SimpleGladeApp):
    def __init__(self, config, gladefile, parent):
        self.is_initializing = True
        SimpleGladeApp.__init__(self, gladefile, 'config_dialog')
        
        self.config_dialog.set_transient_for(parent)
        
        self.config = config
        
        self.fontsel.props.font_name = config.get('font')
        self.cur_font = self.fontsel.props.font_name
        
        self.pprint_chk.props.active = config.get_bool('pprint')
        
        self.reshist_chk.props.active = config.get_bool('use-reshist')
        self.on_reshist_chk_toggled(self.reshist_chk)
        self.reshist_spin.props.value = config.get_int('reshist-size')
        
        self.autofold_chk.props.active = config.get_bool('autofold')
        self.on_autofold_chk_toggled(self.autofold_chk)
        self.autofold_spin.props.value = config.get_int('autofold-numlines')
        
        self.viewer_entry.props.text = eval(config.get('viewer'))
        
        self.autoparen_chk.props.active = config.get_bool('autoparen')
        self.expects_str_entry.props.text = config.get('expects-str-2')
        
        self.vertical_layout_rad.props.active = config.get_bool('vertical-layout')
        self.horizontal_layout_rad.props.active = not config.get_bool('vertical-layout')
        
        self.leave_code_chk.props.active = config.get_bool('leave-code')
        
        self.hide_defs_chk.props.active = config.get_bool('hide-defs')
        
        switch = config.get_bool('matplotlib-ia-switch')
        warn = config.get_bool('matplotlib-ia-warn')
        if switch:
            self.matplotlib_ia_switch_rad.props.active = True
        elif warn:
            self.matplotlib_ia_warn_rad.props.active = True
        else:
            self.matplotlib_ia_ignore_rad.props.active = True
    
        self.ask_on_quit_chk.props.active = config.get_bool('ask-on-quit')
    
        self.themes = dict((name, tags.get_theme(config, name))
                           for name in tags.get_theme_names(config))
        self.cur_theme = self.themes[config.get('current-theme')]

        self.fg_default_rad.set_group(self.fg_special_rad)
        self.bg_default_rad.set_group(self.bg_special_rad)

        TYPE_STRING = gobject.TYPE_STRING
        self.themes_list = gtk.ListStore(TYPE_STRING)
        self.themes_trv.set_model(self.themes_list)
        self.init_themes_list()
        
        # tag, desc, fg, bg
        self.elements_list = gtk.ListStore(TYPE_STRING, TYPE_STRING,
                                           TYPE_STRING, TYPE_STRING)
        self.elements_trv.set_model(self.elements_list)
        # cur_tag is the currently selected tag. It is set to None when props
        # are changed, to mark that they weren't changed as a result of a user
        # action.
        self.cur_tag = None
        self.init_elements_list()

        self.textbuffer = self.textview.get_buffer()
        self.init_textview()

        self.sourcebuffer = gtksourceview2.Buffer()
        self.sourceview = gtksourceview2.View(self.sourcebuffer)
        self.sourcebuffer.set_text(eval(config.get('init-code')))
        self.init_sourceview()
        
        self.font_changed()
        self.theme_changed()
        
        self.is_initializing = False

    def run(self):
        while True:
            r = self.config_dialog.run()
            if r != gtk.RESPONSE_OK:
                return r
            expects_str = self.expects_str_entry.props.text.decode('utf8').strip()
            is_ident = lambda s: re.match(r'[A-Za-z_][A-Za-z0-9_]*$', s)
            if not all(is_ident(s) for s in expects_str.split()):
                warning = _("All names in the auto-quote list must be legal "
                            "Python identifiers and be separated by spaces.")
                msg = gtk.MessageDialog(self.config_dialog, gtk.DIALOG_MODAL,
                                        gtk.MESSAGE_WARNING, gtk.BUTTONS_CLOSE,
                                        warning)
                _response = msg.run()
                msg.destroy()
                self.expects_str_entry.grab_focus()
                continue
            
            # r == gtk.RESPONSE_OK and everything is ok.
            break
        
        config = self.config

        config.set('font', self.fontsel.props.font_name)
        
        config.set_bool('pprint', self.pprint_chk.props.active)
        
        config.set_bool('use-reshist', self.reshist_chk.props.active)
        config.set_int('reshist-size', self.reshist_spin.props.value)
        
        config.set_bool('autofold', self.autofold_chk.props.active)
        config.set_int('autofold-numlines', self.autofold_spin.props.value)
        
        config.set('viewer', repr(self.viewer_entry.props.text.decode('utf8').strip()))
        
        config.set_bool('autoparen', self.autoparen_chk.props.active)
        config.set('expects-str-2', expects_str)
        
        config.set_bool('vertical-layout', self.vertical_layout_rad.props.active)
        
        config.set_bool('leave-code', self.leave_code_chk.props.active)
        
        config.set_bool('hide-defs', self.hide_defs_chk.props.active)
        
        if self.matplotlib_ia_switch_rad.props.active:
            switch = True
            warn = False
        elif self.matplotlib_ia_warn_rad.props.active:
            switch = False
            warn = True
        else:
            switch = warn = False
        config.set_bool('matplotlib-ia-switch', switch)
        config.set_bool('matplotlib-ia-warn', warn)
        
        config.set_bool('ask-on-quit', self.ask_on_quit_chk.props.active)
        
        sb = self.sourcebuffer
        init_code = get_text(sb, sb.get_start_iter(), sb.get_end_iter())
        config.set('init-code', repr(init_code))
        
        tags.remove_themes(config)
        for name, theme in self.themes.iteritems():
            tags.set_theme(config, name, theme)
        cur_theme_name = [name for name, theme in self.themes.iteritems()
                          if theme is self.cur_theme][0]
        config.set('current-theme', cur_theme_name)

        config.save()
        return r # gtk.RESPONSE_OK
    
    def destroy(self):
        self.config_dialog.destroy()
    
    def init_themes_list(self):
        ttv = self.themes_trv; tl = self.themes_list
        for name in self.themes:
            tl.append((name,))
        cr = gtk.CellRendererText()
        cr.props.editable = True
        cr.connect('edited', self.on_theme_renamed)
        ttv.insert_column_with_attributes(0, 'Theme Name', cr, text=0)
        tl.set_sort_column_id(0, gtk.SORT_ASCENDING)
        for i, row in enumerate(tl):
            name = row[0]
            if self.themes[name] is self.cur_theme:
                ttv.set_cursor((i,))
                break
        else:
            assert False, "Didn't find the current theme"
        self.del_theme_btn.props.sensitive = (len(tl) > 1)

    def init_elements_list(self):
        etv = self.elements_trv; el = self.elements_list
        for i, (tag, desc) in enumerate(tags.tag_desc):
            el.insert(i, (tag, desc, None, None))
        letter = gtk.CellRendererText()
        letter.props.text = 'A'
        etv.insert_column_with_attributes(0, 'Preview', letter,
                                          foreground=2, background=3)
        name = gtk.CellRendererText()
        etv.insert_column_with_attributes(1, 'Element Name', name,
                                          text=1)
        etv.set_cursor((0,)) # Default

    def init_textview(self):
        from .tags import (STDIN, STDOUT, STDERR, EXCEPTION, PROMPT, MESSAGE,
                           FOLD_MESSAGE, RESULT_IND, RESULT,
                           KEYWORD, BUILTIN, STRING, NUMBER, COMMENT)
        tb = self.textbuffer
        tags.add_tags(tb)
        def w(s, *tags):
            tb.insert_with_tags_by_name(tb.get_end_iter(), s, *tags)
        w('>>>', PROMPT); w(' '); w('# You can click here', COMMENT); w('\n')
        w('...', PROMPT); w(' '); w('# to choose elements!', COMMENT); w('\n')
        w('...', PROMPT); w(' '); w('def', KEYWORD); w(' add1():\n')
        w('...', PROMPT); w('     num = '); w('input', BUILTIN); w('()\n')
        w('...', PROMPT); w('     '); w('print', KEYWORD); w(' '); \
            w('"What about"', STRING); w(', num+'); w('1', NUMBER); w('\n')
        w('...', PROMPT); w('     '); w('return', KEYWORD); w(' num ** '); \
            w('2', NUMBER); w('\n')
        w('...', PROMPT); w(' add1()\n')
        w('5\n', STDIN)
        w('What about 6\n', STDOUT)
        w('0: ', RESULT_IND); w('25', RESULT); w('\n')
        w('====== New Session ======\n', MESSAGE)
        w('>>>', PROMPT); w(' '); w('from', KEYWORD); w(' sys '); \
            w('import', KEYWORD); w(' stderr\n')
        w('...', PROMPT); w(' '); w('print', KEYWORD); w(' >> stderr, '); \
            w(r'"err\n"', STRING); w(', '); w('1', NUMBER); w('/'); \
            w('0', NUMBER); w('\n')
        w('err\n', STDERR)
        w('Traceback (most recent call last):\n', EXCEPTION)
        w('[About 4 more lines.]', FOLD_MESSAGE)

    def on_textview_realize(self, _widget):
        win = self.textview.get_window(gtk.TEXT_WINDOW_TEXT)
        win.set_cursor(None)
    
    def init_sourceview(self):
        sv = self.sourceview; sb = self.sourcebuffer
        lm = gtksourceview2.LanguageManager()
        python = lm.get_language('python')
        sb.set_language(python)
        self.scrolledwindow_sourceview.add(sv)
        sv.show()

    def font_changed(self):
        # Called when the font was changed, and elements need to be updated
        font = pango.FontDescription(self.cur_font)
        self.textview.modify_font(font)
        self.sourceview.modify_font(font)

    def theme_changed(self):
        # Called when the theme was changed, and elements need to be updated

        theme = self.cur_theme

        tags.apply_theme_text(self.textview, self.textbuffer, theme)
        tags.apply_theme_source(self.sourcebuffer, theme)

        el = self.elements_list
        for i, (tag, _desc) in enumerate(tags.tag_desc):
            el[i][2] = tags.get_actual_color(theme, tag, FG)
            el[i][3] = tags.get_actual_color(theme, tag, BG)
        
        self.update_color_sel_widgets()

    def on_elements_trv_cursor_changed(self, _widget):
        (row,), _col = self.elements_trv.get_cursor()
        self.cur_tag = tags.tag_desc[row][0]
        self.update_color_sel_widgets()

    def update_color_sel_widgets(self):
        tag = self.cur_tag
        # Set cur_tag to None to mark that changes are not the result of user
        # interaction
        self.cur_tag = None

        theme = self.cur_theme
        if tag == DEFAULT:
            self.fg_special_rad.props.active = True
            self.fg_default_rad.props.active = False
            self.fg_default_rad.props.sensitive = False
            self.bg_special_rad.props.active = True
            self.bg_default_rad.props.active = False
            self.bg_default_rad.props.sensitive = False
        else:
            self.fg_special_rad.props.active = theme[tag, FG, ISSET]
            self.fg_default_rad.props.active = not theme[tag, FG, ISSET]
            self.fg_default_rad.props.sensitive = True
            self.bg_special_rad.props.active = theme[tag, BG, ISSET]
            self.bg_default_rad.props.active = not theme[tag, BG, ISSET]
            self.bg_default_rad.props.sensitive = True
        
        self.fg_cbut.props.color = gdk.color_parse(theme[tag, FG, COLOR])
        self.bg_cbut.props.color = gdk.color_parse(theme[tag, BG, COLOR])
        
        self.cur_tag = tag

    def on_viewer_button_clicked(self, _widget):
        def f(filename):
            self.viewer_entry.props.text = filename
        open_dialog(f, _('Choose the viewer program'), self.config_dialog,
                    _('Executables'), '*')
    
    def on_textview_button_press_event(self, _widget, event):
        tv = self.textview
        if tv.get_window(gtk.TEXT_WINDOW_TEXT) is not event.window:
            # Probably a click on the border or something
            return
        x, y = tv.window_to_buffer_coords(gtk.TEXT_WINDOW_TEXT,
                                          int(event.x), int(event.y))
        it = tv.get_iter_at_location(x, y)
        it_tags = it.get_tags()
        if not it_tags:
            tag_index = 0 # Default
        else:
            tag_name = it_tags[-1].props.name
            for i, (tag, _desc) in enumerate(tags.tag_desc):
                if tag == tag_name:
                    tag_index = i
                    break
            else:
                tag_index = 0
        self.elements_trv.set_cursor((tag_index,))

    @staticmethod
    def _format_color(color):
        return '#%04x%04x%04x' % (color.red, color.green, color.blue)
    
    def on_fg_special_rad_toggled(self, _widget):
        is_special = self.fg_special_rad.props.active
        self.fg_cbut.props.sensitive = is_special
        
        if self.cur_tag:
            self.cur_theme[self.cur_tag, FG, ISSET] = is_special
            self.theme_changed()

    def on_fg_cbut_color_set(self, _widget):
        if self.cur_tag:
            color = self._format_color(self.fg_cbut.props.color)
            self.cur_theme[self.cur_tag, FG, COLOR] = color
            self.theme_changed()

    def on_bg_special_rad_toggled(self, _widget):
        is_special = self.bg_special_rad.props.active
        self.bg_cbut.props.sensitive = is_special
        
        if self.cur_tag:
            self.cur_theme[self.cur_tag, BG, ISSET] = is_special
            self.theme_changed()

    def on_bg_cbut_color_set(self, _widget):
        if self.cur_tag:
            color = self._format_color(self.bg_cbut.props.color)
            self.cur_theme[self.cur_tag, BG, COLOR] = color
            self.theme_changed()

    def on_themes_trv_cursor_changed(self, _widget):
        if self.is_initializing:
            return
        ttv = self.themes_trv; tl = self.themes_list
        path, _col = ttv.get_cursor()
        cur_name = tl[path][0]
        self.cur_theme = self.themes[cur_name]
        self.theme_changed()                

    def on_copy_theme_btn_clicked(self, _widget):
        ttv = self.themes_trv; tl = self.themes_list
        path, _col = ttv.get_cursor()
        cur_name = tl[path][0]
        i = 2
        while True:
            new_name = '%s %d' % (cur_name, i)
            if new_name not in self.themes:
                break
            i += 1
        self.themes[new_name] = self.cur_theme = dict(self.themes[cur_name])
        tl.append((new_name,))
        self.del_theme_btn.props.sensitive = True
        tl.set_sort_column_id(0, gtk.SORT_ASCENDING)
        cur_index = [i for i, row in enumerate(tl) if row[0] == new_name][0]
        ttv.set_cursor(cur_index, ttv.get_column(0), start_editing=True)
        self.theme_changed()

    def on_del_theme_btn_clicked(self, _widget):
        self.delete_theme()

    def on_themes_trv_key_press_event(self, _widget, event):
        if gdk.keyval_name(event.keyval) == 'Delete':
            if len(self.themes_list) < 2:
                beep()
            else:
                self.delete_theme()
    
    def delete_theme(self):
        ttv = self.themes_trv; tl = self.themes_list
        assert len(tl) > 1
        path, _col = ttv.get_cursor()
        cur_name = tl[path][0]
        del self.themes[cur_name]
        del tl[path]
        ttv.set_cursor(0)
        self.cur_theme = self.themes[tl[0][0]]
        self.theme_changed()
        if len(tl) < 2:
            self.del_theme_btn.props.sensitive = False

    def on_theme_renamed(self, _widget, path, new_name):
        tl = self.themes_list
        if new_name == tl[path][0]:
            # The name wasn't changed
            return
        if new_name in [row[0] for row in tl]:
            beep()
            return False
        cur_name = tl[path][0]
        theme = self.themes.pop(cur_name)
        self.themes[new_name] = theme
        tl[path][0] = new_name

    def on_notebook_switch_page(self, _widget, _page, _page_num):
        # This should have been on the FontSelection signal, but there isn't
        # one.
        if self.cur_font != self.fontsel.props.font_name:
            self.cur_font = self.fontsel.props.font_name
            self.font_changed()

    def on_reshist_chk_toggled(self, _widget):
        self.reshist_spin.props.sensitive = self.reshist_chk.props.active
    
    def on_autofold_chk_toggled(self, _widget):
        self.autofold_spin.props.sensitive = self.autofold_chk.props.active

    def on_autoparen_chk_toggled(self, _widget):
        self.expects_str_alignment.props.sensitive = self.autoparen_chk.props.active



# Copyright 2010 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

"""
Build library used by the subprocess.
This is not in setup.py so that it may be called at runtime when running
from the source directory.
"""

__all__ = ['build', 'dirs', 'files', 'lib_fns', 'lib_vers']

import sys
import os
from os.path import join, abspath, dirname

try:
    from lib2to3 import refactor
except ImportError:
    py3_available = False
else:
    py3_available = True

dirs = [
    'dreampielib',
    'dreampielib/subprocess',
    'dreampielib/common']

files = [
    'dreampielib/__init__.py',
    'dreampielib/subprocess/__init__.py',
    'dreampielib/subprocess/find_modules.py',
    'dreampielib/subprocess/split_to_singles.py',
    'dreampielib/subprocess/trunc_traceback.py',
    'dreampielib/common/__init__.py',
    'dreampielib/common/objectstream.py',
    'dreampielib/common/brine.py',
    ]

lib_fns = {2: 'subp-py2', 3: 'subp-py3'}
if py3_available:
    lib_vers = [2, 3]
else:
    lib_vers = [2]

def newer(source, target):
    """
    Return True if the source is newer than the target or if the target doesn't
    exist.
    """
    if not os.path.exists(target):
        return True
    
    return os.path.getmtime(source) > os.path.getmtime(target)

class SimpleLogger(object):
    """Used when real logging isn't needed"""
    def debug(self, s):
        pass
    def info(self, s):
        print >> sys.stderr, s
simple_logger = SimpleLogger()

def build(log=simple_logger, force=False):
    dreampielib_dir = dirname(abspath(__file__))
    src_dir = dirname(dreampielib_dir)
    build_dir = join(dreampielib_dir, 'data')
    
    if py3_available:
        avail_fixes = refactor.get_fixers_from_package('lib2to3.fixes')
        rt = refactor.RefactoringTool(avail_fixes)
    
    for ver in lib_vers:
        lib_fn = join(build_dir, lib_fns[ver])

        # Make dirs if they don't exist yet
        if not os.path.exists(lib_fn):
            os.mkdir(lib_fn)
        for dir in dirs:
            dir_fn = join(lib_fn, dir)
            if not os.path.exists(dir_fn):
                os.mkdir(dir_fn)
        
        # Write files if not up to date
        for fn in files:
            src_fn = join(src_dir, fn)
            dst_fn = join(lib_fn, fn)
            if not force and not newer(src_fn, dst_fn):
                continue
            
            if ver == 3:
                log.info("Converting %s to Python 3..." % fn)
            else:
                log.info("Copying %s..." % fn)
            
            f = open(join(src_dir, fn), 'rb')
            src = f.read()
            f.close()
            
            if ver == 3:
                dst = str(rt.refactor_string(src+'\n', fn))[:-1]
            else:
                dst = src
            
            dst = """\
# This file was automatically generated from a file in the source DreamPie dir.
# DO NOT EDIT IT, as your changes will be gone when the file is created again.

""" + dst
            
            f = open(dst_fn, 'wb')
            f.write(dst)
            f.close()



# Copyright 2010 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['History']

from zlib import adler32 as hash_cmd

from .tags import COMMAND, PROMPT
from .common import beep, get_text

# In order to filter out repeating commands, we store the number of times a
# command was encountered. To save memory, we only map the hash of a command
# to a number. The docs say that adler32 is a fast checksum function. I think
# that 32 bits should be enough (you'll get a collision if you have 2**16
# commands, and even that will just mean that a command isn't retreived).
# To ease debugging, uncomment this:
# hash_cmd = lambda s: s

class History(object):
    """
    Manage moving between commands on the text view, and recalling commands
    in the source view.
    """
    def __init__(self, textview, sourceview, sv_changed, config):
        self.textview = textview
        self.textbuffer = textview.get_buffer()
        self.sourceview = sourceview
        self.sourcebuffer = sourceview.get_buffer()
        sv_changed.append(self._on_sv_changed)
        self.recall_1_char_commands = config.get_bool('recall-1-char-commands')

        tb = self.textbuffer

        self.hist_prefix = None
        # Map a command hash to the number of times it has occured in the search.
        # This lets us avoid showing the same command twice.
        self.hist_count = {}
        self.sb_changed = True
        # A handler_id when sb_changed is False.
        self.changed_handler_id = None
        self.hist_mark = tb.create_mark('history', tb.get_end_iter(), False)

    def _on_sv_changed(self, new_sv):
        if self.changed_handler_id:
            self._on_sourcebuffer_changed(None)
        self.sourceview = new_sv
        self.sourcebuffer = new_sv.get_buffer()
    
    def _track_change(self):
        """Set self.sb_changed to False, and add a handler which will set it
        to True on the next change."""
        if not self.sb_changed:
            return
        self.sb_changed = False
        self.changed_handler_id = self.sourcebuffer.connect(
            'changed', self._on_sourcebuffer_changed)
    
    def _on_sourcebuffer_changed(self, _widget):
        self.sb_changed = True
        self.sourcebuffer.disconnect(self.changed_handler_id)
        self.changed_handler_id = None

    def iter_get_command(self, it, only_first_line=False):
        """Get a textiter placed inside (or at the end of) a COMMAND tag.
        Return the text of the tag which doesn't have the PROMPT tag.
        """
        tb = self.textbuffer
        prompt = tb.get_tag_table().lookup(PROMPT)
        command = tb.get_tag_table().lookup(COMMAND)

        it = it.copy()
        if not it.begins_tag(command):
            it.backward_to_tag_toggle(command)
            assert it.begins_tag(command)
        it_end = it.copy(); it_end.forward_to_tag_toggle(command)
        if it.has_tag(prompt):
            it.forward_to_tag_toggle(prompt)
        if it.compare(it_end) >= 0:
            # nothing but prompt
            return ''
        r = []
        while True:
            it2 = it.copy()
            it2.forward_to_tag_toggle(prompt)
            if it2.compare(it_end) >= 0:
                it2 = it.copy()
                it2.forward_to_tag_toggle(command)
                r.append(get_text(tb, it, it2))
                break
            r.append(get_text(tb, it, it2))
            if only_first_line:
                break
            it = it2
            it.forward_to_tag_toggle(prompt)
            if it.compare(it_end) >= 0:
                break
        return ''.join(r)

    def copy_to_sourceview(self):
        # Append the selected command(s) to the sourceview
        tb = self.textbuffer
        sb = self.sourcebuffer
        command = tb.get_tag_table().lookup(COMMAND)

        sel = tb.get_selection_bounds()
        if not sel:
            it = tb.get_iter_at_mark(tb.get_insert())
            if not it.has_tag(command) and not it.ends_tag(command):
                beep()
                return True
            s = self.iter_get_command(it).strip()
        else:
            # Copy all commands which intersect with the selection
            it, end_it = sel
            s = ''
            if it.has_tag(command) or it.ends_tag(command):
                s += self.iter_get_command(it).strip() + '\n'
                if not it.ends_tag(command):
                    it.forward_to_tag_toggle(command)
            assert not it.has_tag(command)
            while True:
                it.forward_to_tag_toggle(command)
                if it.compare(end_it) >= 0:
                    break
                s += self.iter_get_command(it).strip() + '\n'
                it.forward_to_tag_toggle(command)
            s = s.strip()
        if not s:
            beep()
            return True
        cur_text = get_text(sb, sb.get_start_iter(), sb.get_end_iter())
        if cur_text and not cur_text.endswith('\n'):
            s = '\n' + s
        sb.place_cursor(sb.get_end_iter())
        sb.insert_at_cursor(s)
        self.sourceview.scroll_mark_onscreen(sb.get_insert())
        self.sourceview.grab_focus()
        return True

    def history_up(self):
        """Called when the history up command is required"""
        if self.textview.is_focus():
            tb = self.textbuffer
            command = tb.get_tag_table().lookup(COMMAND)
            insert = tb.get_insert()
            it = tb.get_iter_at_mark(insert)
            it.backward_to_tag_toggle(command)
            if it.ends_tag(command):
                it.backward_to_tag_toggle(command)
            self.textbuffer.place_cursor(it)
            self.textview.scroll_mark_onscreen(insert)

        elif self.sourceview.is_focus():
            tb = self.textbuffer
            sb = self.sourcebuffer
            command = tb.get_tag_table().lookup(COMMAND)
            if self.sb_changed:
                if sb.get_end_iter().get_line() != 0:
                    # Don't allow prefixes of more than one line
                    beep()
                    return
                self.hist_prefix = get_text(sb, sb.get_start_iter(),
                                            sb.get_end_iter())
                self.hist_count = {}
                self._track_change()
                tb.move_mark(self.hist_mark, tb.get_end_iter())
            it = tb.get_iter_at_mark(self.hist_mark)
            if it.is_start():
                beep()
                return
            while True:
                it.backward_to_tag_toggle(command)
                if it.ends_tag(command):
                    it.backward_to_tag_toggle(command)
                if not it.begins_tag(command):
                    beep()
                    break
                first_line = self.iter_get_command(it, only_first_line=True).strip()
                if (first_line
                    and first_line.startswith(self.hist_prefix)
                    and (len(first_line) > 2 or self.recall_1_char_commands)):
                    
                    cmd = self.iter_get_command(it).strip()
                    cmd_hash = hash_cmd(cmd)
                    tb.move_mark(self.hist_mark, it)
                    count = self.hist_count.get(cmd_hash, 0) + 1
                    self.hist_count[cmd_hash] = count
                    if count == 1:
                        sb.set_text(cmd)
                        self._track_change()
                        sb.place_cursor(sb.get_end_iter())
                        break
                if it.is_start():
                    beep()
                    return

        else:
            beep()

    def history_down(self):
        """Called when the history down command is required"""
        if self.textview.is_focus():
            tb = self.textbuffer
            command = tb.get_tag_table().lookup(COMMAND)
            insert = tb.get_insert()
            it = tb.get_iter_at_mark(insert)
            it.forward_to_tag_toggle(command)
            if it.ends_tag(command):
                it.forward_to_tag_toggle(command)
            self.textbuffer.place_cursor(it)
            self.textview.scroll_mark_onscreen(insert)

        elif self.sourceview.is_focus():
            tb = self.textbuffer
            sb = self.sourcebuffer
            command = tb.get_tag_table().lookup(COMMAND)
            if self.sb_changed:
                beep()
                return
            it = tb.get_iter_at_mark(self.hist_mark)
            passed_one = False
            while True:
                if not it.begins_tag(command):
                    # Return the source buffer to the prefix and everything
                    # to initial state.
                    sb.set_text(self.hist_prefix)
                    sb.place_cursor(sb.get_end_iter())
                    # Since we change the text and not called _track_change,
                    # it's like the user did it and hist_prefix is not longer
                    # meaningful.
                    break
                first_line = self.iter_get_command(it, only_first_line=True).strip()
                if (first_line
                    and first_line.startswith(self.hist_prefix)
                    and (len(first_line) > 2 or self.recall_1_char_commands)):
                    
                    cmd = self.iter_get_command(it).strip()
                    cmd_hash = hash_cmd(cmd)
                    tb.move_mark(self.hist_mark, it)
                    if self.hist_count[cmd_hash] == 1:
                        if passed_one:
                            sb.set_text(cmd)
                            self._track_change()
                            sb.place_cursor(sb.get_end_iter())
                            break
                        else:
                            passed_one = True
                    self.hist_count[cmd_hash] -= 1

                it.forward_to_tag_toggle(command)
                it.forward_to_tag_toggle(command)
                

        else:
            beep()



# This file was automatically generated from a file in the source DreamPie dir.
# DO NOT EDIT IT, as your changes will be gone when the file is created again.

# Copyright 2009 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

"""
Send objects over a socket by brining them.
"""

__all__ = ['send_object', 'recv_object']

import sys
py3k = (sys.version_info[0] == 3)
import struct

# This was "from . import brine", but a bug in 2to3 in Python 2.6.5
# converted it to "from .. import brine", so I changed that.
from ..common import brine

if not py3k:
    empty_bytes = ''
else:
    empty_bytes = bytes()

def send_object(sock, obj):
    """Send an object over a socket"""
    s = brine.dump(obj)
    msg = struct.pack('<l', len(s)) + s
    sock.sendall(msg)

def recv_object(sock):
    """Receive an object over a socket"""
    length_str = empty_bytes
    while len(length_str) < 4:
        r = sock.recv(4 - len(length_str))
        if not r:
            raise IOError("Socket closed unexpectedly")
        length_str += r
    length, = struct.unpack('<i', length_str)
    parts = []
    len_received = 0
    while len_received < length:
        r = sock.recv(length - len_received)
        if not r:
            raise IOError("Socket closed unexpectedly")
        parts.append(r)
        len_received += len(r)
    s = empty_bytes.join(parts)
    obj = brine.load(s)
    return obj



# Copyright 2009 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['make_keyhandler_decorator', 'handle_keypress',
           'parse_keypress_event']

"""
Help handling keypress events. The functions here should be used like this:

keyhandlers = {}
keyhandler = make_keyhandler_decorator(keyhandlers)
class Whatever:
    @keyhandler('Return', 0)
    def on_return(self):
        # Do something
    def on_keypress(self, widget, event):
        handle_keypress(self, event, keyhandlers)
"""

from gtk import gdk

# We ignore all other mods. There isn't a standard modifier for Alt,
# and we don't use it anyway in our shortcuts.
handled_mods = gdk.SHIFT_MASK | gdk.CONTROL_MASK

def make_keyhandler_decorator(keyhandlers_dict):
    def keyhandler(keyval, state):
        def decorator(func):
            keyhandlers_dict[keyval, state] = func
            return func
        return decorator
    return keyhandler

def parse_keypress_event(event):
    """
    Get a keypress event, return a tuple of (keyval_name, state).
    Will return (None, None) when no appropriate tuple is available.
    """
    r = gdk.keymap_get_default().translate_keyboard_state(
        event.hardware_keycode, event.state, event.group)
    if r is None:
        # This seems to be the case when pressing CapsLock on win32
        return (None, None)
    keyval, _group, _level, consumed_mods = r
    state = event.state & ~consumed_mods & handled_mods
    keyval_name = gdk.keyval_name(keyval)
    if keyval_name == 'KP_Enter':
        keyval_name = 'Return'
    return keyval_name, state

def handle_keypress(self, event, keyhandlers_dict):
    keyval_name, state = parse_keypress_event(event)
    try:
        func = keyhandlers_dict[keyval_name, state]
    except KeyError:
        pass
    else:
        return func(self)




# Copyright 2010 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

import sys
import os
from os import path
import time
import tempfile
from optparse import OptionParser
import subprocess
import webbrowser
import re
from keyword import iskeyword

import logging
from logging import debug
#logging.basicConfig(format="dreampie: %(message)s", level=logging.DEBUG)

def find_data_dir():
    """
    Find the data directory in which to find files.
    If we are inside the source directory, build subp zips.
    """
    # The data directory is normally located at dreampielib/data.
    # When running under py2exe, it is in the same directory as the executable.
    # Running inside the source directory is detected by the presence of a
    # file called 'dreampie' in the same directory as 'dreampielib'.
    from os.path import join, dirname, abspath, isfile

    if hasattr(sys, 'frozen'):
        return abspath(join(dirname(sys.executable), 'data'))
    dreampielib_dir = dirname(dirname(abspath(__file__)))
    if isfile(join(dirname(dreampielib_dir), 'dreampie')):
        # We're in the source path. Build zips if needed, and return the right
        # dir.
        from ..subp_lib import build
        build()
    return join(dreampielib_dir, 'data')

data_dir = find_data_dir()
gladefile = path.join(data_dir, 'dreampie.glade')

def load_pygtk():
    """On win32, load PyGTK from subdirectory, if available."""
    from os.path import join, dirname, abspath
    
    if hasattr(sys, 'frozen'):
        pygtk_dir = join(dirname(abspath(sys.executable)), 'gtk-2.0')
    else:
        pygtk_dir = join(dirname(dirname(dirname(abspath(__file__)))), 'gtk-2.0')
    
    if os.path.isdir(pygtk_dir):
        sys.path.insert(0, pygtk_dir)
        import runtime #@UnresolvedImport

if sys.platform == 'win32':
    load_pygtk()

import gobject
gobject.threads_init() #@UndefinedVariable
import gtk
from gtk import gdk, glade
import pango
import gtksourceview2
from . import gtkexcepthook
gtkexcepthook.install(gladefile)

try:
    from glib import timeout_add, idle_add
except ImportError:
    # In PyGObject 2.14, it's in gobject.
    from gobject import timeout_add, idle_add

from .. import __version__

from .SimpleGladeApp import SimpleGladeApp
from .keyhandler import (make_keyhandler_decorator, handle_keypress,
                         parse_keypress_event)
from .config import Config
from .config_dialog import ConfigDialog
from .write_command import write_command
from .newline_and_indent import newline_and_indent
from .output import Output
from .folding import Folding
from .selection import Selection
from .status_bar import StatusBar
from .vadj_to_bottom import VAdjToBottom
from .history import History
from .hist_persist import HistPersist
from .autocomplete import Autocomplete
from .call_tips import CallTips
from .autoparen import Autoparen
from .crash_workaround import TextViewCrashWorkaround
from .subprocess_handler import SubprocessHandler, StartError
from .common import beep, get_text, TimeoutError
from .file_dialogs import save_dialog
from .tags import (OUTPUT, STDIN, STDOUT, STDERR, EXCEPTION, PROMPT, COMMAND,
                   COMMAND_DEFS, COMMAND_SEP, MESSAGE, RESULT_IND, RESULT)
from . import tags
from .update_check import update_check
from . import bug_report

INDENT_WIDTH = 4

# Default line length, by which we set the default window size
LINE_LEN = 80

# Time to wait before autocompleting, to see if the user continues to type
AUTOCOMPLETE_WAIT = 400

# Time to wait for the subprocess for a result. The subprocess may be doing
# idle jobs, and so not return a result.
SUBP_WAIT_TIMEOUT_S = .5

# Maybe someday we'll want translations...
_ = lambda s: s

# A decorator for managing sourceview key handlers
sourceview_keyhandlers = {}
sourceview_keyhandler = make_keyhandler_decorator(sourceview_keyhandlers)

def get_widget(name):
    """Create a widget from the glade file."""
    xml = glade.XML(gladefile, name)
    return xml.get_widget(name)

class DreamPie(SimpleGladeApp):
    def __init__(self, pyexec, runfile):
        """
        pyexec - the Python interpreter executable
        runfile - a filename to run upon startup, or None.
        """
        SimpleGladeApp.__init__(self, gladefile, 'window_main')
        self.load_popup_menus()
        self.set_mac_accelerators()
        
        self.config = Config()

        self.window_main.set_icon_from_file(
            path.join(data_dir, 'dreampie.png'))

        self.textbuffer = tb = self.textview.get_buffer()
        self.init_textbufferview()
        # Mark where the cursor was when the popup menu was popped
        self.popup_mark = tb.create_mark('popup-mark', tb.get_start_iter(),
                                         left_gravity=True)

        # Remove the page in the notebook, which was added because empty
        # notebooks cause warnings
        self.notebook.remove_page(0)
        # A list of callbacks to call when changing the sourcebuffer
        self.sv_changed = []
        self.sourceview = self.create_sourcebufferview()
        self.sourcebuffer = self.sourceview.get_buffer()
        # A tuple (page_num, text) of the recently closed tab
        self.reopen_tab_data = None
        
        # last (font, vertical_layout) configured. If they are changed,
        # configure() will resize the window and place the paned.
        self.last_configured_layout = (None, None)
        self.configure()

        self.output = Output(self.textview)
        
        self.folding = Folding(self.textbuffer, LINE_LEN)

        self.selection = Selection(self.textview, self.sourceview,
                                   self.sv_changed,
                                   self.on_is_something_selected_changed)

        self.status_bar = StatusBar(self.sourcebuffer, self.sv_changed,
                                    self.statusbar)

        self.vadj_to_bottom = VAdjToBottom(self.scrolledwindow_textview
                                           .get_vadjustment())

        self.history = History(self.textview, self.sourceview, self.sv_changed,
                               self.config)

        self.recent_manager = gtk.recent_manager_get_default()
        self.menuitem_recent = [self.menuitem_recent0, self.menuitem_recent1,
                                self.menuitem_recent2, self.menuitem_recent3]
        self.recent_filenames = [None] * len(self.menuitem_recent)
        self.recent_manager.connect('changed', self.on_recent_manager_changed)

        self.histpersist = HistPersist(self.window_main, self.textview,
                                       self.status_bar, self.recent_manager)
        self.update_recent()
        
        self.autocomplete = Autocomplete(self.sourceview,
                                         self.sv_changed,
                                         self.window_main,
                                         self.complete_attributes,
                                         self.complete_firstlevels,
                                         self.get_func_args,
                                         self.find_modules,
                                         self.get_module_members,
                                         self.complete_filenames,
                                         self.complete_dict_keys,
                                         INDENT_WIDTH)
        
        # Hack: we connect this signal here, so that it will have lower
        # priority than the key-press event of autocomplete, when active.
        self.sourceview_keypress_handler = self.sourceview.connect(
            'key-press-event', self.on_sourceview_keypress)
        self.sv_changed.append(self.on_sv_changed)
        
        self.call_tips = CallTips(self.sourceview, self.sv_changed,
                                  self.window_main, self.get_func_doc,
                                  INDENT_WIDTH)
        
        self.autoparen = Autoparen(self.sourcebuffer, self.sv_changed,
                                   self.is_callable_only,
                                   self.get_expects_str,
                                   self.autoparen_show_call_tip,
                                   INDENT_WIDTH)

        self.subp = SubprocessHandler(
            pyexec, data_dir,
            self.on_stdout_recv, self.on_stderr_recv, self.on_object_recv,
            self.on_subp_terminated)
        # Number of RPC calls that timed out and expecting results
        self._n_unclaimed_results = 0
        try:
            self.subp.start()
        except StartError, e:
            msg = gtk.MessageDialog(
                None, gtk.DIALOG_MODAL, gtk.MESSAGE_ERROR, gtk.BUTTONS_CLOSE,
                _("Couldn't start subprocess: %s") % e)
            _response = msg.run()
            msg.destroy()
            print >> sys.stderr, e
            sys.exit(1)

        # Is the subprocess executing a command
        self.set_is_executing(False)
        
        # Are we trying to shut down
        self.is_terminating = False

        self.window_main.show()
        
        self.subp_welcome, self.subp_can_mask_sigint = (
            self.call_subp(u'get_subprocess_info'))
        self.show_welcome()
        self.configure_subp()
        self.run_init_code(runfile)
        bug_report.set_subp_info(pyexec, self.subp_welcome)

        if self.config.get_bool('show-getting-started'):
            self.show_getting_started_dialog()
            self.config.set_bool('show-getting-started', False)
            self.config.save()
        
        update_check(self.on_update_available)
        
    def on_sv_changed(self, new_sv):
        self.sourceview.disconnect(self.sourceview_keypress_handler)
        self.sourceview = new_sv
        self.sourcebuffer = new_sv.get_buffer()
        self.sourceview_keypress_handler = self.sourceview.connect(
            'key-press-event', self.on_sourceview_keypress)
    
    def load_popup_menus(self):
        # Load popup menus from the glade file. Would not have been needed if
        # popup menus could be children of windows.
        xml = glade.XML(gladefile, 'popup_sel_menu')
        xml.signal_autoconnect(self)
        self.popup_sel_menu = xml.get_widget('popup_sel_menu')
        
        xml = glade.XML(gladefile, 'popup_nosel_menu')
        xml.signal_autoconnect(self)
        self.popup_nosel_menu = xml.get_widget('popup_nosel_menu')
        self.fold_unfold_section_menu = xml.get_widget('fold_unfold_section_menu')
        self.copy_section_menu = xml.get_widget('copy_section_menu')
        self.view_section_menu = xml.get_widget('view_section_menu')
        self.save_section_menu = xml.get_widget('save_section_menu')
    
    def set_mac_accelerators(self):
        # Set up accelerators suitable for the Mac.
        # Ctrl-Up and Ctrl-Down are taken by the window manager, so we use
        # Ctrl-PgUp and Ctrl-PgDn.
        # We want it to be easy to switch, so both sets of keys are always
        # active, but only one, most suitable for each platform, is displayed
        # in the menu.
        
        accel_group = gtk.accel_groups_from_object(self.window_main)[0]
        menu_up = self.menuitem_history_up
        UP = gdk.keyval_from_name('Up')
        PGUP = gdk.keyval_from_name('Prior')
        menu_dn = self.menuitem_history_down
        DN = gdk.keyval_from_name('Down')
        PGDN = gdk.keyval_from_name('Next')

        if sys.platform != 'darwin':
            menu_up.add_accelerator('activate', accel_group, PGUP,
                                    gdk.CONTROL_MASK, 0)
            menu_dn.add_accelerator('activate', accel_group, PGDN,
                                    gdk.CONTROL_MASK, 0)
        else:
            menu_up.remove_accelerator(accel_group, UP, gdk.CONTROL_MASK)
            menu_up.add_accelerator('activate', accel_group, PGUP,
                                    gdk.CONTROL_MASK, gtk.ACCEL_VISIBLE)
            menu_up.add_accelerator('activate', accel_group, UP,
                                    gdk.CONTROL_MASK, 0)
    
            menu_dn.remove_accelerator(accel_group, DN, gdk.CONTROL_MASK)
            menu_dn.add_accelerator('activate', accel_group, PGDN,
                                    gdk.CONTROL_MASK, gtk.ACCEL_VISIBLE)
            menu_dn.add_accelerator('activate', accel_group, DN,
                                    gdk.CONTROL_MASK, 0)

    def on_cut(self, _widget):
        return self.selection.cut()

    def on_copy(self, _widget):
        return self.selection.copy()

    def on_copy_commands_only(self, _widget):
        return self.selection.copy_commands_only()

    def on_paste(self, _widget):
        return self.selection.paste()

    def on_is_something_selected_changed(self, is_something_selected):
        self.menuitem_cut.props.sensitive = is_something_selected
        self.menuitem_copy.props.sensitive = is_something_selected
        self.menuitem_copy_commands_only.props.sensitive = is_something_selected
        self.menuitem_interrupt.props.sensitive = not is_something_selected

    # Source buffer, Text buffer

    def init_textbufferview(self):
        tv = self.textview
        tb = self.textbuffer

        tv.set_wrap_mode(gtk.WRAP_CHAR)

        self.textview_crash_workaround = TextViewCrashWorkaround(tv)

        tags.add_tags(tb)

        tv.connect('key-press-event', self.on_textview_keypress)
        tv.connect('focus-in-event', self.on_textview_focus_in)

    def get_char_width_height(self):
        tv = self.textview
        context = tv.get_pango_context()
        metrics = context.get_metrics(tv.style.font_desc,
                                      context.get_language())
        charwidth = pango.PIXELS(metrics.get_approximate_digit_width())
        # I don't know why +1
        charheight = pango.PIXELS(metrics.get_ascent() + metrics.get_descent())+1
        return charwidth, charheight
    
    def set_window_size(self, vertical_layout):
        charwidth, charheight = self.get_char_width_height()
        if vertical_layout:
            # I don't know why I have to add 2, but it works.
            width = charwidth*(LINE_LEN+2)
            height = charheight*30
        else:
            width = charwidth*((LINE_LEN-10)*2+2)
            height = charheight*26
        self.window_main.resize(width, height)
        
        # Set the position of the paned. We wait until it is exposed because
        # then its max_position is meaningful.
        # In vertical layout we set it to maximum, since the sourceview has
        # a minimum height.
        def callback(_widget, _event):
            if vertical_layout:
                pane = self.vpaned_main
                pane.set_position(pane.props.max_position)
            else:
                pane = self.hpaned_main
                pane.set_position(pane.props.max_position // 2)
            self.sourceview.disconnect(callback_id)
        callback_id = self.sourceview.connect('expose-event', callback)

    def create_sourcebufferview(self, page_num=None):
        sb = gtksourceview2.Buffer()
        sv = gtksourceview2.View(sb)
        sv.show()
        sv.connect('focus-in-event', self.on_sourceview_focus_in)
        _charwidth, charheight = self.get_char_width_height()
        self.configure_sourceview(sv)

        lm = gtksourceview2.LanguageManager()
        lm.set_search_path([path.join(data_dir, 'language-specs')])
        sb.set_language(lm.get_language('python'))
        scroll = gtk.ScrolledWindow()
        scroll.show()
        scroll.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        scroll.add(sv)
        scroll.set_size_request(-1, charheight * 4)
        
        lbl = gtk.Label('      ')
        if page_num is None:
            page_num = self.notebook.get_current_page() + 1
        self.notebook.insert_page(scroll, lbl, page_num)
        self.notebook.set_current_page(page_num)
        sv.grab_focus()

        return sv

    def sv_scroll_cursor_onscreen(self):
        self.sourceview.scroll_mark_onscreen(self.sourcebuffer.get_insert())

    def on_textview_focus_in(self, _widget, _event):
        # Clear the selection of the sourcebuffer
        self.sourcebuffer.move_mark(self.sourcebuffer.get_selection_bound(),
                                    self.sourcebuffer.get_iter_at_mark(
                                        self.sourcebuffer.get_insert()))

    def on_sourceview_focus_in(self, _widget, _event):
        # Clear the selection of the textbuffer
        self.textbuffer.move_mark(self.textbuffer.get_selection_bound(),
                                  self.textbuffer.get_iter_at_mark(
                                        self.textbuffer.get_insert()))

    def write(self, data, *tag_names):
        self.textbuffer.insert_with_tags_by_name(
            self.textbuffer.get_end_iter(), data, *tag_names)

    def write_output(self, data, tag_names, onnewline=False, addbreaks=True):
        """
        Call self.output.write with the given arguments, and autofold if needed.
        """
        it = self.output.write(data, tag_names, onnewline, addbreaks)
        if self.config.get_bool('autofold'):
            self.folding.autofold(it, self.config.get_int('autofold-numlines'))
    
    def set_is_executing(self, is_executing):
        self.is_executing = is_executing
        label = _(u'Execute Code') if not is_executing else _(u'Write Input')
        self.menuitem_execute.child.props.label = label
        self.menuitem_discard_hist.props.sensitive = not is_executing

    @staticmethod
    def replace_gtk_quotes(source):
        # Work around GTK+ bug https://bugzilla.gnome.org/show_bug.cgi?id=610928
        # in order to fix bug #525469 - replace fancy quotes with regular
        # quotes.
        return source.replace(u'\xa8', '"').replace(u'\xb4', "'")
    
    def execute_source(self):
        """Execute the source in the source buffer.
        """
        sb = self.sourcebuffer
        source = get_text(sb, sb.get_start_iter(), sb.get_end_iter())
        source = source.rstrip()
        source = self.replace_gtk_quotes(source)
        try:
            # There's a chance that the subprocess won't reply, because it's
            # busy doing "idle jobs". For most queries we can ask for an answer
            # and if it doesn't arrive, cancel what we tried to do and ignore
            # the answer when it comes. However, here we can't let the execute
            # function run and ignore the result, so we first call 'pause_idle'.
            # If we don't get a reply for pause_idle, we don't execute.
            self.call_subp_noblock(u'pause_idle')
        except TimeoutError:
            self.subp.send_object((u'resume_idle', ()))
            self._n_unclaimed_results += 1

            self.status_bar.set_status(_("The subprocess is currently busy"))
            beep()
            return
            
        is_ok, syntax_error_info = self.call_subp(u'execute', source)
        if not is_ok:
            if syntax_error_info:
                msg, lineno, offset = syntax_error_info
                status_msg = _("Syntax error: %s (at line %d col %d)") % (
                    msg, lineno+1, offset+1)
                # Work around a bug: offset may be wrong, which will cause
                # gtk to crash if using sb.get_iter_at_line_offset.
                iter = sb.get_iter_at_line(lineno)
                iter.forward_chars(offset+1)
                sb.place_cursor(iter)
            else:
                # Incomplete
                status_msg = _("Command is incomplete")
                sb.place_cursor(sb.get_end_iter())
            self.status_bar.set_status(status_msg)
            beep()
        else:
            write_command(self.write, source.strip())
            self.output.start_new_section()
            if not self.config.get_bool('leave-code'):
                sb.delete(sb.get_start_iter(), sb.get_end_iter())
            self.vadj_to_bottom.scroll_to_bottom()
            self.set_is_executing(True)

    def send_stdin(self):
        """Send the contents of the sourcebuffer as stdin."""
        sb = self.sourcebuffer
        s = get_text(sb, sb.get_start_iter(), sb.get_end_iter())
        if not s.endswith('\n'):
            s += '\n'

        self.write_output(s, [COMMAND, STDIN], addbreaks=False)
        self.write('\r', COMMAND_SEP)
        self.output.start_new_section()
        self.vadj_to_bottom.scroll_to_bottom()

        if not self.config.get_bool('leave-code'):
            sb.delete(sb.get_start_iter(), sb.get_end_iter())

        self.subp.write(s)

    @sourceview_keyhandler('Return', 0)
    def on_sourceview_return(self):
        sb = self.sourcebuffer

        # If we are on the first line, and it doesn't end with a ' ':
        #   * If we are not executing, try to execute (if failed, continue
        #     with normal behavior)
        #   * If we are executing, send the line as stdin.
        insert_iter = sb.get_iter_at_mark(sb.get_insert())
        if (insert_iter.equal(sb.get_end_iter())
            and insert_iter.get_line() == 0
            and insert_iter.get_offset() != 0
            and not get_text(sb, sb.get_start_iter(),
                             insert_iter).endswith(' ')):

            if not self.is_executing:
                source = get_text(sb, sb.get_start_iter(), sb.get_end_iter())
                source = source.rstrip()
                source = self.replace_gtk_quotes(source)
                try:
                    is_incomplete = self.call_subp_noblock(u'is_incomplete', source)
                except TimeoutError:
                    is_incomplete = True
                if not is_incomplete:
                    self.execute_source()
                    return True
            else:
                # is_executing
                self.send_stdin()
                return True
                
        # If we are after too many newlines, the user probably just wanted to
        # execute - notify him.
        # We check if this line is empty and the previous one is.
        show_execution_tip = False
        if insert_iter.equal(sb.get_end_iter()):
            it = sb.get_end_iter()
            # This goes to the beginning of the line, and another line
            # backwards, so we get two lines
            it.backward_lines(1)
            text = get_text(sb, it, sb.get_end_iter())
            if not text.strip():
                show_execution_tip = True

        # We didn't execute, so newline-and-indent.
        r = newline_and_indent(self.sourceview, INDENT_WIDTH)

        if show_execution_tip:
            self.status_bar.set_status(_(
                "Tip: To execute your code, use Ctrl+Enter."))
        return r

    @sourceview_keyhandler('Tab', 0)
    def on_sourceview_tab(self):
        sb = self.sourcebuffer
        sel = sb.get_selection_bounds()
        if not sel:
            insert = sb.get_iter_at_mark(sb.get_insert())
            insert_linestart = sb.get_iter_at_line(insert.get_line())
            line = get_text(sb, insert_linestart, insert)
    
            if not line.strip():
                # We are at the beginning of a line, so indent - forward to next
                # "tab stop"
                sb.insert_at_cursor(' '*(INDENT_WIDTH - len(line)%INDENT_WIDTH))
    
            else:
                # Completion should come here
                self.autocomplete.show_completions(is_auto=False, complete=True)
        else:
            # Indent
            start, end = sel
            start = sb.get_iter_at_line(start.get_line())
            if not end.ends_line():
                end.forward_to_line_end()
            text = get_text(sb, start, end)
            newtext = '\n'.join('    '+line for line in text.split('\n'))
            start_offset = start.get_offset()
            sb.delete(start, end)
            sb.insert(end, newtext)
            sb.select_range(sb.get_iter_at_offset(start_offset), end)

        self.sv_scroll_cursor_onscreen()
        return True

    @sourceview_keyhandler('ISO_Left_Tab', 0)
    def on_sourceview_shift_tab(self):
        sb = self.sourcebuffer
        sel = sb.get_selection_bounds()
        if sel:
            start, end = sel
        else:
            start = end = sb.get_iter_at_mark(sb.get_insert())
        start = sb.get_iter_at_line(start.get_line())
        if not end.ends_line():
            end.forward_to_line_end()
        text = get_text(sb, start, end)
        lines = text.split('\n')
        if not all(line.startswith('    ')
                   for line in lines if line.strip() != ''):
            beep()
        else:
            newlines = [line[4:] for line in lines]
            newtext = '\n'.join(newlines)
            start_offset = start.get_offset()
            sb.delete(start, end)
            sb.insert(end, newtext)
            sb.select_range(sb.get_iter_at_offset(start_offset), end)
        return True

    @sourceview_keyhandler('Home', 0)
    def on_sourceview_home(self):
        # If the cursor is already at the beginning of the line, move to the
        # beginning of the text.
        sb = self.sourcebuffer
        insert = sb.get_iter_at_mark(sb.get_insert())
        if insert.starts_line():
            while insert.get_char() == ' ':
                insert.forward_char()
            sb.place_cursor(insert)
            return True
    
    @sourceview_keyhandler('BackSpace', 0)
    def on_sourceview_backspace(self):
        sb = self.sourcebuffer
        insert = sb.get_iter_at_mark(sb.get_insert())
        insert_linestart = sb.get_iter_at_line(insert.get_line())
        line = get_text(sb, insert_linestart, insert)

        if line and not line.strip():
            # There are only space before us, so remove spaces up to last
            # "tab stop"
            delete_from = ((len(line) - 1) // INDENT_WIDTH) * INDENT_WIDTH
            it = sb.get_iter_at_line_offset(insert.get_line(), delete_from)
            sb.delete(it, insert)
            self.sv_scroll_cursor_onscreen()
            return True

        return False

    # The following 3 handlers are for characters which may trigger automatic
    # opening of the completion list. (slash and backslash depend on path.sep)
    # We leave the final decision whether to open the list to the autocompleter.
    # We just notify it that the char was inserted and the user waited a while.

    @sourceview_keyhandler('period', 0)
    def on_sourceview_period(self):
        timeout_add(AUTOCOMPLETE_WAIT, self.check_autocomplete, '.')
    @sourceview_keyhandler('slash', 0)
    def on_sourceview_slash(self):
        timeout_add(AUTOCOMPLETE_WAIT, self.check_autocomplete, '/')
    @sourceview_keyhandler('backslash', 0)
    def on_sourceview_backslash(self):
        timeout_add(AUTOCOMPLETE_WAIT, self.check_autocomplete, '\\')
    @sourceview_keyhandler('bracketleft', 0)
    def on_sourceview_bracketleft(self):
        timeout_add(AUTOCOMPLETE_WAIT, self.check_autocomplete, '[')

    def check_autocomplete(self, last_char):
        """
        If the last char in the sourcebuffer is last_char, call
        show_completions.
        """
        sb = self.sourcebuffer
        if self.sourceview.is_focus():
            it = sb.get_iter_at_mark(sb.get_insert())
            it2 = it.copy()
            it2.backward_chars(1)
            char = get_text(sb, it2, it)
            if char == last_char:
                self.autocomplete.show_completions(is_auto=True, complete=False)
        # return False so as not to be called repeatedly.
        return False

    @sourceview_keyhandler('parenleft', 0)
    def on_sourceview_parenleft(self):
        idle_add(self.call_tips.show, True)

    def on_sourceview_keypress(self, _widget, event):
        return handle_keypress(self, event, sourceview_keyhandlers)


    # Autoparen
    
    @sourceview_keyhandler('space', 0)
    def on_sourceview_space(self):
        """
        If a space was hit after a callable-only object, add parentheses.
        """
        if self.is_executing:
            return False
        if not self.config.get_bool('autoparen'):
            return False
        
        return self.autoparen.add_parens()

    def is_callable_only(self, expr):
        return self.call_subp_catch(u'is_callable_only', expr)
    
    def get_expects_str(self):
        return set(self.config.get('expects-str-2').split())
    
    def autoparen_show_call_tip(self):
        self.call_tips.show(is_auto=True)


    # History

    def on_textview_keypress(self, _widget, event):
        keyval_name, state = parse_keypress_event(event)
        if (keyval_name, state) == ('Return', 0):
            return self.history.copy_to_sourceview()

    def on_history_up(self, _widget):
        self.history.history_up()

    def on_history_down(self, _widget):
        self.history.history_down()

    # Subprocess

    def show_welcome(self):
        s = self.subp_welcome + 'DreamPie %s\n' % __version__
        self.write(s, MESSAGE)
        self.output.start_new_section()

    def configure_subp(self):
        config = self.config
        
        if config.get_bool('use-reshist'):
            reshist_size = config.get_int('reshist-size')
        else:
            reshist_size = 0
        self.call_subp(u'set_reshist_size', reshist_size)
        self.menuitem_clear_reshist.props.sensitive = (reshist_size > 0)
        
        self.call_subp(u'set_pprint', config.get_bool('pprint'))
        
        self.call_subp(u'set_matplotlib_ia',
                       config.get_bool('matplotlib-ia-switch'),
                       config.get_bool('matplotlib-ia-warn'))
        
    def run_init_code(self, runfile=None):
        """
        Runs the init code.
        This will result in the code being run and a '>>>' printed afterwards.
        If there's no init code, will just print '>>>'.
        
        If runfile is given, will also execute the code in that.
        """
        init_code = unicode(eval(self.config.get('init-code')))
        if runfile:
            msg = "Running %s" % runfile
            # This should be both valid py3 and py2 code.
            init_code += ('\n\nprint(%r)\nexec(open(%r).read())\n'
                          % (msg, runfile))
        if init_code:
            is_ok, syntax_error_info = self.call_subp(u'execute', init_code)
            if not is_ok:
                msg, lineno, offset = syntax_error_info
                warning = _(
                    "Could not run initialization code because of a syntax "
                    "error:\n"
                    "%s at line %d col %d.") % (msg, lineno+1, offset+1)
                msg = gtk.MessageDialog(self.window_main, gtk.DIALOG_MODAL,
                                        gtk.MESSAGE_WARNING, gtk.BUTTONS_CLOSE,
                                        warning)
                _response = msg.run()
                msg.destroy()
            else:
                self.set_is_executing(True)
        if not self.is_executing:
            self.write('>>> ', COMMAND, PROMPT)

    def on_subp_terminated(self):
        if self.is_terminating:
            return
        # This may raise an exception if subprocess couldn't be started,
        # but hopefully if it was started once it will be started again.
        self._n_unclaimed_results = 0
        self.subp.start()
        self.set_is_executing(False)
        self.write('\n')
        self.write(
            '==================== New Session ====================\n',
            MESSAGE)
        self.output.start_new_section()
        self.configure_subp()
        self.run_init_code()
        self.vadj_to_bottom.scroll_to_bottom()
        self.sourceview.grab_focus()

    def on_restart_subprocess(self, _widget):
        self.subp.kill()

    def on_stdout_recv(self, data):
        self.write_output(data, STDOUT)

    def on_stderr_recv(self, data):
        self.write_output(data, STDERR)

    def call_subp(self, funcname, *args):
        """
        Make an RPC call, blocking until an answer is received.
        """
        assert not self.is_executing
        
        while self._n_unclaimed_results:
            self.subp.recv_object()
            
        self.subp.send_object((funcname, args))
        return self.subp.recv_object()

    def call_subp_noblock(self, funcname, *args):
        """
        Make a non-blocking RPC call.
        Will wait for SUBP_WAIT_TIMEOUT_S and if no answer is received will
        raise a TimeoutError. The query will be executed when the subprocess
        becomes responsive again, but will be discarded.
        """
        assert not self.is_executing
        
        while self._n_unclaimed_results:
            returned = self.subp.wait_for_object(SUBP_WAIT_TIMEOUT_S)
            if returned:
                self.subp.recv_object()
            else:
                raise TimeoutError
            
        self.subp.send_object((funcname, args))
        returned = self.subp.wait_for_object(SUBP_WAIT_TIMEOUT_S)
        if returned:
            return self.subp.recv_object()
        else:
            self._n_unclaimed_results += 1
            raise TimeoutError

    def call_subp_catch(self, funcname, *args):
        """
        Make a non-blocking RPC call.
        If executing, return None.
        If a TimeoutError is raised, catch it and return None.
        """
        if self.is_executing:
            return None
        try:
            return self.call_subp_noblock(funcname, *args)
        except TimeoutError:
            return None
    
    def on_object_recv(self, obj):
        if self._n_unclaimed_results:
            self._n_unclaimed_results -= 1
            return
        
        assert self.is_executing

        is_success, val_no, val_str, exception_string, rem_stdin = obj

        if not is_success:
            self.write_output(exception_string, EXCEPTION, onnewline=True)
        else:
            if val_str is not None:
                if val_no is not None:
                    self.write_output('%d: ' % val_no, RESULT_IND,
                                      onnewline=True)
                self.write_output(val_str+'\n', RESULT)
        self.write('>>> ', COMMAND, PROMPT)
        self.set_is_executing(False)
        self.handle_rem_stdin(rem_stdin)

    def handle_rem_stdin(self, rem_stdin):
        """
        Add the stdin text that was not processed to the source buffer.
        Remove it from the text buffer (we check that the STDIN text is
        consistent with rem_stdin - otherwise we give up)
        """
        if not rem_stdin:
            return

        self.sourcebuffer.insert(self.sourcebuffer.get_start_iter(), rem_stdin)
        self.sv_scroll_cursor_onscreen()

        tb = self.textbuffer
        stdin = tb.get_tag_table().lookup(STDIN)
        it = tb.get_end_iter()
        if not it.ends_tag(stdin):
            it.backward_to_tag_toggle(stdin)
        while True:
            it2 = it.copy()
            it2.backward_to_tag_toggle(stdin)
            cur_stdin = get_text(tb, it2, it, True)
            min_len = min(len(cur_stdin), len(rem_stdin))
            assert min_len > 0
            if cur_stdin[-min_len:] != rem_stdin[-min_len:]:
                debug("rem_stdin doesn't match what's in textview")
                break
            it2.forward_chars(len(cur_stdin)-min_len)
            tb.delete(it2, it)
            rem_stdin = rem_stdin[:-min_len]
            if not rem_stdin:
                break
            else:
                it = it2
                # if rem_stdin is left, it2 must be at the beginning of the
                # stdin region.
                it2.backward_to_tag_toggle(stdin)
                assert it2.ends_tag(stdin)

    def on_execute_command(self, _widget):
        if self.is_executing:
            self.send_stdin()
        elif self.sourcebuffer.get_char_count() == 0:
            beep()
        else:
            self.execute_source()
        return True

    def on_interrupt(self, _widget):
        if self.subp_can_mask_sigint or self.is_executing:
            self.subp.interrupt()
        else:
            self.status_bar.set_status(
                _("A command isn't being executed currently"))
            beep()

    # History persistence
    
    def on_save_history(self, _widget):
        self.histpersist.save()
    
    def on_save_history_as(self, _widget):
        self.histpersist.save_as()
    
    def on_load_history(self, _widget):
        self.histpersist.load()
    
    # Recent history files
    
    def on_recent_manager_changed(self, _recent_manager):
        self.update_recent()
    
    def update_recent(self):
        """Update the menu and self.recent_filenames"""
        rman = self.recent_manager
        recent_items = [it for it in rman.get_items()
                        if it.has_application('dreampie')
                        and it.get_uri().startswith('file://')]
        # it.get_visited() makes more sense, but since we call RecentManager.add
        # when we open and when we save, get_modified() does the trick.
        recent_items.sort(key=lambda it: it.get_modified(),
                          reverse=True)
        self.menuitem_recentsep.props.visible = (len(recent_items) > 0)
        for i, menuitem in enumerate(self.menuitem_recent):
            if i < len(recent_items):
                it = recent_items[i]
                fn = it.get_uri()[len('file://'):]
                menuitem.props.visible = True
                menuitem.child.props.label = "_%d %s" % (i, fn)
                self.recent_filenames[i] = fn
            else:
                menuitem.props.visible = False
                self.recent_filenames[i] = None
    
    def on_menuitem_recent(self, widget):
        num = self.menuitem_recent.index(widget)
        fn = self.recent_filenames[num]
        self.histpersist.load_filename(fn)
    
    # Discard history
    
    def discard_hist_before_tag(self, tag):
        """
        Discard history before the given tag. If tag == COMMAND, this discards
        all history, and if tag == MESSAGE, this discards previous sessions.
        """
        tb = self.textbuffer
        
        tag = tb.get_tag_table().lookup(tag)
        it = tb.get_end_iter()
        it.backward_to_tag_toggle(tag)
        if not it.begins_tag(tag):
            it.backward_to_tag_toggle(tag)
        tb.delete(tb.get_start_iter(), it)

    def on_discard_history(self, _widget):
        xml = glade.XML(gladefile, 'discard_hist_dialog')
        d = xml.get_widget('discard_hist_dialog')
        d.set_transient_for(self.window_main)
        d.set_default_response(gtk.RESPONSE_OK)

        previous_rad = xml.get_widget('previous_rad')
        all_rad = xml.get_widget('all_rad')
        previous_rad.set_group(all_rad)
        previous_rad.props.active = True
        
        r = d.run()
        d.destroy()
        
        if r == gtk.RESPONSE_OK:
            tb = self.textbuffer
            if previous_rad.props.active:
                self.discard_hist_before_tag(MESSAGE)
            else:
                self.discard_hist_before_tag(COMMAND)
                tb.insert_with_tags_by_name(
                    tb.get_start_iter(),
                    '================= History Discarded =================\n',
                    MESSAGE)
            self.status_bar.set_status(_('History discarded.'))
            self.histpersist.forget_filename()

    # Folding
    
    def on_section_menu_activate(self, widget):
        """
        Called when the used clicked a section-related item in a popup menu.
        """
        tb = self.textbuffer
        it = tb.get_iter_at_mark(self.popup_mark)
        r = self.folding.get_section_status(it)
        if r is None:
            # May happen if something was changed in the textbuffer between
            # popup and activation
            return
        typ, is_folded, start_it = r
        
        if widget is self.fold_unfold_section_menu:
            # Fold/Unfold
            if is_folded is None:
                # No point in folding.
                beep()
            elif not is_folded:
                self.folding.fold(typ, start_it)
            else:
                self.folding.unfold(typ, start_it)
        else:
            if typ == COMMAND:
                text = self.history.iter_get_command(start_it)
            else:
                end_it = start_it.copy()
                end_it.forward_to_tag_toggle(self.folding.get_tag(typ))
                text = get_text(tb, start_it, end_it)
            if sys.platform == 'win32':
                text = text.replace('\n', '\r\n')
            if widget is self.copy_section_menu:
                # Copy
                self.selection.clipboard.set_text(text)
            elif widget is self.view_section_menu:
                # View
                fd, fn = tempfile.mkstemp()
                os.write(fd, text)
                os.close(fd)
                viewer = eval(self.config.get('viewer'))
                self.spawn_and_forget('%s %s' % (viewer, fn))
            elif widget is self.save_section_menu:
                # Save
                def func(filename):
                    f = open(filename, 'wb')
                    f.write(text)
                    f.close()
                save_dialog(func, _("Choose where to save the section"),
                            self.main_widget, _("All Files"), "*", None)
            else:
                assert False, "Unexpected widget"
            
    def spawn_and_forget(self, argv):
        """
        Start a process and forget about it.
        """
        if sys.platform == 'linux2':
            # We use a trick so as not to create zombie processes: we fork,
            # and let the fork spawn the process (actually another fork). The
            # (first) fork immediately exists, so the process we spawned is
            # made the child of process number 1.
            pid = os.fork()
            if pid == 0:
                _p = subprocess.Popen(argv, shell=True)
                os._exit(0)
            else:
                os.waitpid(pid, 0)
        else:
            _p = subprocess.Popen(argv, shell=True)
    
    def on_double_click(self, event):
        """If we are on a folded section, unfold it and return True, to
        avoid event propagation."""
        tv = self.textview

        if tv.get_window(gtk.TEXT_WINDOW_TEXT) is not event.window:
            # Probably a click on the border or something
            return
        x, y = tv.window_to_buffer_coords(gtk.TEXT_WINDOW_TEXT,
                                          int(event.x), int(event.y))
        it = tv.get_iter_at_location(x, y)
        r = self.folding.get_section_status(it)
        if r is not None:
            typ, is_folded, start_it = r
            if is_folded:
                self.folding.unfold(typ, start_it)
                return True
    
    def on_fold_last(self, _widget):
        self.folding.fold_last()
    
    def on_unfold_last(self, _widget):
        self.folding.unfold_last()

    # Notebook tabs
    
    def on_notebook_switch_page(self, _widget, _page, page_num):
        new_sv = self.notebook.get_nth_page(page_num).get_child()
        for cb in self.sv_changed:
            cb(new_sv)
    
    def new_tab(self, index=None):
        # The following line should result in on_notebook_switch_page, which
        # will take care of calling on_sv_change functions.
        self.create_sourcebufferview(index)
        self.notebook.props.show_tabs = True
        self.reopen_tab_data = None
        self.menuitem_reopen_tab.props.sensitive = False
    
    def on_new_tab(self, _widget):
        self.new_tab()
    
    def on_reopen_tab(self, _widget):
        index, text = self.reopen_tab_data
        self.new_tab(index)
        self.sourcebuffer.set_text(text)
    
    def on_close_tab(self, _widget):
        if self.notebook.get_n_pages() == 1:
            beep()
            return
        else:
            self.close_current_tab()
    
    def close_current_tab(self):
        assert self.notebook.get_n_pages() > 1
        cur_page = self.notebook.get_current_page()
        text = get_text(self.sourcebuffer, self.sourcebuffer.get_start_iter(),
                        self.sourcebuffer.get_end_iter())
        if text:
            self.reopen_tab_data = (cur_page, text)
            self.menuitem_reopen_tab.props.sensitive = True
        else:
            self.reopen_tab_data = None
            self.menuitem_reopen_tab.props.sensitive = False

        scrolledwin = self.notebook.get_nth_page(cur_page)
        new_page = cur_page-1 if cur_page > 0 else 1
        # This should result in on_notebook_switch_page which will set
        # everything to use the new sourcebuffer
        self.notebook.set_current_page(new_page)
        assert self.sourceview is not scrolledwin.get_child()
        self.notebook.remove_page(cur_page)
        if self.notebook.get_n_pages() == 1:
            self.notebook.props.show_tabs = False
        if True:
            scrolledwin.destroy()
        else:
            # Verify that the sourceview and sourcebuffer are indeed destroyed,
            # and not referenced anywhere
            import weakref, gc
            r = weakref.ref(scrolledwin.get_child().get_buffer())
            scrolledwin.destroy()
            gc.collect()
            assert r() is None
    
    def on_prev_tab(self, _widget):
        self.notebook.prev_page()
    
    def on_next_tab(self, _widget):
        self.notebook.next_page()
    
    # Other events

    def on_show_completions(self, _widget):
        self.autocomplete.show_completions(is_auto=False, complete=False)

    def complete_dict_keys(self, expr):
        return self.call_subp_catch(u'complete_dict_keys', expr)

    def complete_attributes(self, expr):
        return self.call_subp_catch(u'complete_attributes', expr)

    def complete_firstlevels(self):
        return self.call_subp_catch(u'complete_firstlevels')
    
    def get_func_args(self, expr):
        return self.call_subp_catch(u'get_func_args', expr)
    
    def find_modules(self, expr):
        return self.call_subp_catch(u'find_modules', expr)
    
    def get_module_members(self, expr):
        return self.call_subp_catch(u'get_module_members', expr)
    
    def complete_filenames(self, str_prefix, text, str_char, add_quote):
        return self.call_subp_catch(u'complete_filenames', str_prefix, text, str_char,
                                    add_quote)

    def on_show_calltip(self, _widget):
        self.call_tips.show(is_auto=False)

    def get_func_doc(self, expr):
        return self.call_subp_catch(u'get_func_doc', expr)

    def configure(self):
        """
        Apply configuration. Called on initialization and after configuration
        was changed by the configuration dialog.
        """
        config = self.config
        tv = self.textview; tb = self.textbuffer
        sourceviews = [self.notebook.get_nth_page(i).get_child()
                       for i in range(self.notebook.get_n_pages())]
        
        font_name = config.get('font')
        font = pango.FontDescription(font_name)
        tv.modify_font(font)
        for sv in sourceviews:
            sv.modify_font(font)

        theme = tags.get_theme(self.config, self.config.get('current-theme'))
        tags.apply_theme_text(tv, tb, theme)
        for sv in sourceviews:
            tags.apply_theme_source(sv.get_buffer(), theme)

        vertical_layout = self.config.get_bool('vertical-layout')
        if vertical_layout:
            pane = self.vpaned_main; other_pane = self.hpaned_main
            self.notebook.props.tab_pos = gtk.POS_BOTTOM
        else:
            pane = self.hpaned_main; other_pane = self.vpaned_main
            self.notebook.props.tab_pos = gtk.POS_TOP
        pane.props.visible = True
        other_pane.props.visible = False
        if pane.get_child1() is None:
            child1 = other_pane.get_child1(); other_pane.remove(child1)
            child2 = other_pane.get_child2(); other_pane.remove(child2)
            pane.pack1(child1, resize=True, shrink=False)
            pane.pack2(child2, resize=not vertical_layout, shrink=False)
        
        # If the fonts were changed, we might need to enlarge the window
        last_font, last_vertical = self.last_configured_layout
        if last_font != font or last_vertical != vertical_layout:
            self.set_window_size(vertical_layout)
            self.last_configured_layout = font, vertical_layout
        
        command_defs = self.textbuffer.get_tag_table().lookup(COMMAND_DEFS)
        command_defs.props.invisible = config.get_bool('hide-defs')

    def configure_sourceview(self, sv):
        """
        Apply configuration to a newly created sourceview.
        This does the same for a single sourceview as configure() does for
        all of them.
        """
        font_name = self.config.get('font')
        font = pango.FontDescription(font_name)
        sv.modify_font(font)
        theme = tags.get_theme(self.config, self.config.get('current-theme'))
        tags.apply_theme_source(sv.get_buffer(), theme)
    
    def on_preferences(self, _widget):
        cd = ConfigDialog(self.config, gladefile, self.window_main)
        r = cd.run()
        if r == gtk.RESPONSE_OK:
            self.configure()
            self.configure_subp()
        cd.destroy()

    def on_clear_reshist(self, _widget):
        try:
            self.call_subp_noblock(u'clear_reshist')
        except TimeoutError:
            # Will happen anyway when idle job ends
            pass
        self.status_bar.set_status(_("Result history cleared."))

    def on_close(self, _widget, _event):
        self.quit()
        return True

    def on_quit(self, _widget):
        self.quit()

    def quit(self):
        was_saved = self.histpersist.was_saved()
        if (self.textbuffer.get_modified()
            and (was_saved or self.config.get_bool('ask-on-quit'))):
            d = gtk.MessageDialog(
                parent=self.window_main,
                flags=gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                type=gtk.MESSAGE_WARNING,
                message_format=_('Save history before closing?'))
            d.props.secondary_text = _("If you don't save, your history will be lost.")
            CANCEL, DISCARD, SAVE = range(3)
            discard_btn = d.add_button(_("Close _without saving"), DISCARD)
            _cancel_btn = d.add_button(_("_Cancel"), CANCEL)
            save_btn = d.add_button(_("_Save"), SAVE)
            if not was_saved:
                dontask_chk = gtk.CheckButton(
                    _("Don't ask me again when the history was never saved"))
                dontask_chk.show()
                d.get_content_area().pack_start(dontask_chk, fill=True, expand=False)
                d.set_default_response(DISCARD)
                discard_btn.grab_focus()
            else:
                d.set_default_response(SAVE)
                save_btn.grab_focus()

            r = d.run()
            if not was_saved and r == DISCARD and dontask_chk.props.active:
                self.config.set_bool('ask-on-quit', False)
                self.config.save()
            if r == SAVE:
                saved = self.histpersist.save()
                quit = saved
            elif r == DISCARD:
                quit = True
            else:
                quit = False
            d.destroy()
        else:
            quit = True
        if quit:
            self.is_terminating = True
            self.window_main.destroy()
            self.subp.kill()
            gtk.main_quit()

    def on_about(self, _widget):
        d = get_widget('about_dialog')
        d.set_transient_for(self.window_main)
        d.set_version(__version__)
        d.set_logo(gdk.pixbuf_new_from_file(
            path.join(data_dir, 'dreampie.png')))
        d.run()
        d.destroy()
    
    def on_update_available(self, is_git, latest_name, latest_time):
        date = time.strftime('%Y/%m/%d', time.localtime(latest_time))
        if is_git:
            msg = _("A new git commit is available, from %s. "
                    "Run 'git pull' to update." % date)
        else:
            self.get_update_menu.show()
            msg = _("A new DreamPie version, %s, is available. "
                    "Click Help->Get New Version to update." % latest_name)
        self.status_bar.set_status(msg)
    
    def on_get_update_menu_activate(self, _widget):
        webbrowser.open('http://www.dreampie.org/download.html')
    
    def on_report_bug(self, _widget):
        bug_report.bug_report(self.window_main, gladefile, None)
    
    def on_homepage(self, _widget):
        webbrowser.open('http://www.dreampie.org/')
    
    def on_getting_started(self, _widget):
        self.show_getting_started_dialog()
    
    def show_getting_started_dialog(self):
        d = get_widget('getting_started_dialog')
        d.set_transient_for(self.window_main)
        d.run()
        d.destroy()
    
    def on_textview_button_press_event(self, _widget, event):
        if event.button == 3:
            self.show_popup_menu(event)
            return True
        
        if event.type == gdk._2BUTTON_PRESS:
            return self.on_double_click(event)
    
    def show_popup_menu(self, event):
        tv = self.textview
        tb = self.textbuffer
        
        if tb.get_has_selection():
            self.popup_sel_menu.popup(None, None, None, event.button,
                                      event.get_time())
        else:
            if tv.get_window(gtk.TEXT_WINDOW_TEXT) is not event.window:
                # Probably a click on the border or something
                return
            x, y = tv.window_to_buffer_coords(gtk.TEXT_WINDOW_TEXT,
                                              int(event.x), int(event.y))
            it = tv.get_iter_at_location(x, y)
            r = self.folding.get_section_status(it)
            if r is not None:
                typ, is_folded, _start_it = r
                if typ == OUTPUT:
                    typ_s = _('Output Section')
                else:
                    typ_s = _('Code Section')
                self.fold_unfold_section_menu.props.visible = (
                    is_folded is not None)
                self.fold_unfold_section_menu.child.props.label = (
                    _('Unfold %s') if is_folded else _('Fold %s')) % typ_s
                self.copy_section_menu.child.props.label = _('Copy %s') % typ_s
                self.view_section_menu.child.props.label = _('View %s') % typ_s
                self.save_section_menu.child.props.label = _('Save %s') % typ_s
                self.view_section_menu.props.visible = \
                    bool(eval(self.config.get('viewer')))
                
                tb.move_mark(self.popup_mark, it)
                self.popup_nosel_menu.popup(None, None, None, event.button,
                                            event.get_time())
            else:
                beep()

def main():
    usage = "%prog [options] [python-executable]"
    version = 'DreamPie %s' % __version__
    parser = OptionParser(usage=usage, version=version)
    parser.add_option("--run", dest="runfile",
                      help="A file to run upon initialization. It will be "
                      "run only once.")
    if sys.platform == 'win32':
        parser.add_option("--hide-console-window", action="store_true",
                          dest="hide_console",
                          help="Hide the console window")

    opts, args = parser.parse_args()
    
    if len(args) > 1:
        parser.error("Can accept at most one argument")
    if len(args) == 1:
        pyexec = args[0]
    elif 'dreampie' in sys.executable.lower():
        # We are under py2exe.
        msg = gtk.MessageDialog(
            None, gtk.DIALOG_MODAL, gtk.MESSAGE_ERROR, gtk.BUTTONS_CLOSE,
            _("DreamPie must be given the file name of a Python interpreter. "
              "Please create a shortcut to something like '%s "
              "--hide-console-window c:\\python26\\python.exe'.")
            % os.path.abspath(sys.argv[0]))
        _response = msg.run()
        msg.destroy()
        sys.exit(1)
    else:
        pyexec = sys.executable
        
    
    if sys.platform == 'win32' and opts.hide_console:
        from .hide_console_window import hide_console_window
        hide_console_window()

    gtk.widget_set_default_direction(gtk.TEXT_DIR_LTR)
    _dp = DreamPie(pyexec, opts.runfile)
    gtk.main()



# Copyright 2009 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['Config']

import sys
import os
from ConfigParser import RawConfigParser
from StringIO import StringIO

from .odict import OrderedDict

# We use expects-str-2, because expects-str had a different format (uses repr)
# in DreamPie 1.1

default_config = """
[DreamPie]
show-getting-started = True
font=Courier New 10
current-theme = Dark
pprint = True
use-reshist = True
reshist-size = 30
autofold = True
autofold-numlines = 30
viewer = ''
init-code = ''
autoparen = True
expects-str-2 = execfile chdir open run runeval
vertical-layout = True
ask-on-quit = True
matplotlib-ia-switch = False
matplotlib-ia-warn = True

recall-1-char-commands = False
hide-defs = False
leave-code = False

[Dark theme]
is-active = True

default-fg = white
default-bg = black

stdin-fg = white
stdin-bg = black
stdout-fg = #bcffff
stdout-bg = black
stderr-fg = #ff8080
stderr-bg = black
result-ind-fg = blue
result-ind-bg = black
result-fg = #bcffff
result-bg = black
exception-fg = #ff8080
exception-bg = black
prompt-fg = #e400b6
prompt-bg = black
message-fg = yellow
message-bg = black
fold-message-fg = #a7a7a7
fold-message-bg = #003b6c

keyword-fg = #ff7700
keyword-bg = black
builtin-fg = #efcfcf
builtin-bg = black
string-fg = #00e400
string-bg = black
number-fg = #aeacff
number-bg = black
comment-fg = #c9a3a0
comment-bg = black

bracket-match-fg = white
bracket-match-bg = darkblue
bracket-1-fg = #abffab
bracket-1-bg = black
bracket-2-fg = #dfabff
bracket-2-bg = black
bracket-3-fg = #ffabab
bracket-3-bg = black
error-fg = white
error-bg = red

stdin-fg-set = False
stdin-bg-set = False
stdout-fg-set = True
stdout-bg-set = False
stderr-fg-set = True
stderr-bg-set = False
result-ind-fg-set = True
result-ind-bg-set = False
result-fg-set = True
result-bg-set = False
exception-fg-set = True
exception-bg-set = False
prompt-fg-set = True
prompt-bg-set = False
message-fg-set = True
message-bg-set = False
fold-message-fg-set = True
fold-message-bg-set = True

keyword-fg-set = True
keyword-bg-set = False
builtin-fg-set = True
builtin-bg-set = False
string-fg-set = True
string-bg-set = False
number-fg-set = True
number-bg-set = False
comment-fg-set = True
comment-bg-set = False

bracket-match-fg-set = False
bracket-match-bg-set = True
bracket-1-fg-set = True
bracket-1-bg-set = False
bracket-2-fg-set = True
bracket-2-bg-set = False
bracket-3-fg-set = True
bracket-3-bg-set = False
error-fg-set = False
error-bg-set = True

[Light theme]
is-active = True

default-fg = black
default-bg = white

stdin-fg = #770000
stdin-bg = white
stdout-fg = blue
stdout-bg = white
stderr-fg = red
stderr-bg = white
result-ind-fg = #808080
result-ind-bg = white
result-fg = blue
result-bg = white
exception-fg = red
exception-bg = white
prompt-fg = #770000
prompt-bg = white
message-fg = #008000
message-bg = white
fold-message-fg = #404040
fold-message-bg = #b2ddff

keyword-fg = #ff7700
keyword-bg = white
builtin-fg = #0000ff
builtin-bg = white
string-fg = #00aa00
string-bg = white
number-fg = blue
number-bg = white
comment-fg = #dd0000
comment-bg = white

bracket-match-fg = black
bracket-match-bg = lightblue
bracket-1-fg = #005400
bracket-1-bg = white
bracket-2-fg = #9400f0
bracket-2-bg = white
bracket-3-fg = brown
bracket-3-bg = #a50000
error-fg = black
error-bg = red

stdin-fg-set = False
stdin-bg-set = False
stdout-fg-set = True
stdout-bg-set = False
stderr-fg-set = True
stderr-bg-set = False
result-ind-fg-set = True
result-ind-bg-set = False
result-fg-set = True
result-bg-set = False
exception-fg-set = True
exception-bg-set = False
prompt-fg-set = True
prompt-bg-set = False
message-fg-set = True
message-bg-set = False
fold-message-fg-set = True
fold-message-bg-set = True

keyword-fg-set = True
keyword-bg-set = False
builtin-fg-set = True
builtin-bg-set = False
string-fg-set = True
string-bg-set = False
number-fg-set = True
number-bg-set = False
comment-fg-set = True
comment-bg-set = False

bracket-match-fg-set = False
bracket-match-bg-set = True
bracket-1-fg-set = True
bracket-1-bg-set = False
bracket-2-fg-set = True
bracket-2-bg-set = False
bracket-3-fg-set = True
bracket-3-bg-set = False
error-fg-set = False
error-bg-set = True

"""

def get_config_fn():
    if sys.platform != 'win32':
        return os.path.expanduser('~/.dreampie')
    else:
        # On win32, expanduser doesn't work when the path includes unicode
        # chars.
        import ctypes
        MAX_PATH = 255
        nFolder = 26 # CSIDL_APPDATA
        flags = 0
        buf = ctypes.create_unicode_buffer(MAX_PATH)
        ctypes.windll.shell32.SHGetFolderPathW(None, nFolder, None, flags, buf)
        return os.path.join(buf.value, 'DreamPie')

class Config(object):
    """
    Manage configuration - a simple wrapper around RawConfigParser.
    Upon initialization, the loaded file is updated with the default values.
    config.save() will save the current state.
    """
    def __init__(self):
        self.filename = get_config_fn()
        try:
            self.parser = RawConfigParser(dict_type=OrderedDict)
        except TypeError:
            # Python versions < 2.6 don't support dict_type
            self.parser = RawConfigParser()
        f = StringIO(default_config)
        self.parser.readfp(f)
        self.parser.read(self.filename)
        self.save()
    
    def get(self, key, section='DreamPie'):
        return self.parser.get(section, key)
    
    def get_bool(self, key, section='DreamPie'):
        return self.parser.getboolean(section, key)
    
    def get_int(self, key, section='DreamPie'):
        return self.parser.getint(section, key)
    
    def set(self, key, value, section='DreamPie'):
        self.parser.set(section, key, value)
    
    def set_bool(self, key, value, section='DreamPie'):
        value_str = 'True' if value else 'False'
        self.set(key, value_str, section)
    
    def set_int(self, key, value, section='DreamPie'):
        if value != int(value):
            raise ValueError("Expected an int, got %r" % value)
        self.set(key, '%d' % value, section)
    
    def sections(self):
        return self.parser.sections()

    def has_section(self, section):
        return self.parser.has_section(section)

    def add_section(self, section):
        return self.parser.add_section(section)

    def remove_section(self, section):
        return self.parser.remove_section(section)

    def save(self):
        f = open(self.filename, 'w')
        self.parser.write(f)
        f.close()




# Copyright 2012 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['TextViewCrashWorkaround']

import sys
import ctypes as ct

import gtk
import glib

class TextViewCrashWorkaround(object):
    """
    This class fixes the annoying crash which happens when the mouse hovers
    over a folded area which is updated - bug #525429.
    As of 2012/02, a patch to the underlying GTK bug was submitted.
    The problem is that when updating a gtk.TextView, it leaves some processing
    to be done when idle. However, if an event occurs (such as the mouse moving
    over the widget), GTK first handles the event and then processes the idle
    job. The event is handled when the textview is in inconsistent state, and
    we get a crash.
    
    The solution is to listen to the event, and let it propagate only after the
    idle job was done. To do this, we check a (semi) private field of the
    TextView instance, which has the handle of the idle callback, and process
    GTK events until is it zeroed.
    
    The place in the struct is hardcoded, but it seems that it has never changed
    in GTK+-2. Just to be on the safer side, we check - if the field doesn't
    change to zero after GTK handled the events, we print an error and don't
    try to fix again.
    """
    
    # Offset in bytes to the first_validate_idle field in the GtkTextView struct
    first_validate_idle_offset = 192
    
    def __init__(self, textview):
        if gtk.gtk_version[0] != 2:
            return
        self._wrong_offset = False
        textview.connect('event', self._on_textview_event)
    
    def _on_textview_event(self, textview, _event):
        if self._wrong_offset:
            return
        first_validate_idle = ct.c_uint.from_address(
            hash(textview)+self.first_validate_idle_offset)
        con = glib.main_context_default()
        while first_validate_idle.value != 0:
            if not con.pending():
                # No pending callbacks, and still not 0? We have the wrong offset
                self._wrong_offset = True
                print >> sys.stderr, 'Warning: wrong first_validate_idle offset'
                return
            con.iteration()



# Copyright 2010 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['Autocomplete']

import string
import re

from .hyper_parser import HyperParser
from .autocomplete_window import AutocompleteWindow, find_prefix_range
from .common import beep, get_text

# This string includes all chars that may be in an identifier
ID_CHARS = string.ascii_letters + string.digits + "_"
ID_CHARS_DOT = ID_CHARS + '.'

class Autocomplete(object):
    def __init__(self, sourceview, sv_changed, window_main,
                 complete_attributes, complete_firstlevels, get_func_args,
                 find_modules, get_module_members, complete_filenames,
                 complete_dict_keys,
                 INDENT_WIDTH):
        self.sourceview = sourceview
        sv_changed.append(self._on_sv_changed)
        self.complete_attributes = complete_attributes
        self.complete_firstlevels = complete_firstlevels
        self.get_func_args = get_func_args
        self.find_modules = find_modules
        self.get_module_members = get_module_members
        self.complete_filenames = complete_filenames
        self.complete_dict_keys = complete_dict_keys
        self.INDENT_WIDTH = INDENT_WIDTH

        self.window = AutocompleteWindow(sourceview, sv_changed, window_main,
                                         self._on_complete)

    def _on_sv_changed(self, new_sv):
        self.sourceview = new_sv
    
    def show_completions(self, is_auto, complete):
        """
        If complete is False, just show the completion list.
        If complete is True, complete as far as possible. If there's only
        one completion, don't show the window.

        If is_auto is True, don't beep if can't find completions.
        """
        sb = self.sourceview.get_buffer()
        text = get_text(sb, sb.get_start_iter(), sb.get_end_iter())
        index = sb.get_iter_at_mark(sb.get_insert()).get_offset()
        hp = HyperParser(text[:index], index, self.INDENT_WIDTH)

        if hp.is_in_code():
            line = text[text.rfind('\n', 0, index)+1:index].lstrip()
            if line.startswith('import '):
                res = self._complete_modules(line, is_auto)
            elif line.startswith('from '):
                if len((line+'x').split()) == 3:
                    # The third word should be "import".
                    res = self._complete_import(line)
                elif ' import ' not in line:            
                    res = self._complete_modules(line, is_auto)
                else:
                    res = self._complete_module_members(line, is_auto)
            elif line.endswith('['):
                # We complete dict keys either after a '[' or in a string
                # after a '['.
                res = self._complete_dict_keys(text, index, hp, is_auto)
            else:
                res = self._complete_attributes(text, index, hp, is_auto)
        elif hp.is_in_string():
            if text[max(hp.bracketing[hp.indexbracket][0]-1,0)] == '[':
                res = self._complete_dict_keys(text, index, hp, is_auto)
            else:
                res = self._complete_filenames(text, index, hp, is_auto)
        else:
            # Not in string and not in code
            res = None

        if res is not None:
            comp_prefix, public, private, is_case_insen = res
        else:
            if not is_auto:
                beep()
            return

        combined = public + private
        if is_case_insen:
            combined.sort(key = lambda s: s.lower())
            combined_keys = [s.lower() for s in combined]
        else:
            combined.sort()
            combined_keys = combined
        comp_prefix_key = comp_prefix.lower() if is_case_insen else comp_prefix
        start, end = find_prefix_range(combined_keys, comp_prefix_key)
        if start == end:
            # No completions
            if not is_auto:
                beep()
            return

        if complete:
            # Find maximum prefix
            first = combined_keys[start]
            last = combined_keys[end-1]
            i = 0
            while i < len(first) and i < len(last) and first[i] == last[i]:
                i += 1
            if i > len(comp_prefix):
                sb.insert_at_cursor(combined[start][len(comp_prefix):i])
                comp_prefix = first[:i]
            if end == start + 1:
                # Only one matching completion - don't show the window
                self._on_complete()
                return

        self.window.show(public, private, is_case_insen, len(comp_prefix))
        
    def _complete_dict_keys(self, text, index, hp, is_auto):
        """
        Return (comp_prefix, public, private, is_case_insen) 
        (string, list, list, bool).
        If shouldn't complete - return None.
        """
        # Check whether auto-completion is really appropriate,
        if is_auto and text[index-1] != '[':
            return
        
        is_in_code = hp.is_in_code()
        opener, _closer = hp.get_surrounding_brackets('[')
        if opener is None:
            return
        hp.set_index(opener)
        comp_what = hp.get_expression()
        if not comp_what:
            # It's not an index, but a list - complete as if the '[' wasn't there.
            hp.set_index(index)
            if is_in_code:
                return self._complete_attributes(text, index, hp, is_auto)
            else:
                return self._complete_filenames(text, index, hp, is_auto)
        if is_auto and '(' in comp_what:
            # Don't evaluate expressions which may contain a function call.
            return
        key_reprs = self.complete_dict_keys(comp_what)
        if key_reprs is None:
            return
        if text[index:index+1] != ']':
            key_reprs = [x+']' for x in key_reprs]

        comp_prefix = text[opener+1:index]
        public = key_reprs
        private = []
        is_case_insen = False
        return (comp_prefix, public, private, is_case_insen)

    def _complete_attributes(self, text, index, hp, is_auto):
        """
        Return (comp_prefix, public, private, is_case_insen) 
        (string, list, list, bool).
        If shouldn't complete - return None.
        """
        # Check whether autocompletion is really appropriate
        if is_auto and text[index-1] != '.':
            return
        
        i = index
        while i and text[i-1] in ID_CHARS:
            i -= 1
        comp_prefix = text[i:index]
        if i and text[i-1] == '.':
            hp.set_index(i-1)
            comp_what = hp.get_expression()
            if not comp_what:
                return
            if is_auto and '(' in comp_what:
                # Don't evaluate expressions which may contain a function call.
                return
            public_and_private = self.complete_attributes(comp_what)
            if public_and_private is None: # The subprocess is busy
                return
            public, private = public_and_private
        else:
            public_and_private = self.complete_firstlevels()
            if public_and_private is None: # The subprocess is busy
                return
            public, private = public_and_private
            
            # If we are inside a function call after a ',' or '(',
            # get argument names.
            if text[:i].rstrip()[-1:] in (',', '('):
                opener, _closer = hp.get_surrounding_brackets('(')
                if opener:
                    hp.set_index(opener)
                    expr = hp.get_expression()
                    if expr and '(' not in expr:
                        # Don't need to execute a function just to get arguments
                        args = self.get_func_args(expr)
                        if args is not None:
                            public.extend(args)
                            public.sort()
        
        is_case_insen = False
        return comp_prefix, public, private, is_case_insen

    def _complete_import(self, line):
        """
        Complete the word "import"...
        """
        i = len(line)
        while i and line[i-1] in ID_CHARS:
            i -= 1
        comp_prefix = line[i:]
        public = ['import']
        private = []
        is_case_insen = False
        return comp_prefix, public, private, is_case_insen
        
    
    def _complete_modules(self, line, is_auto):
        """
        line - the stripped line from its beginning to the cursor.
        Return (comp_prefix, public, private, is_case_insen) 
        (string, list, list, bool).
        If shouldn't complete - return None.
        """
        # Check whether autocompletion is really appropriate
        if is_auto and line[-1] != '.':
            return
        
        i = len(line)
        while i and line[i-1] in ID_CHARS:
            i -= 1
        comp_prefix = line[i:]
        if i and line[i-1] == '.':
            i -= 1
            j = i
            while j and line[j-1] in ID_CHARS_DOT:
                j -= 1
            comp_what = line[j:i]
        else:
            comp_what = u''
        
        modules = self.find_modules(comp_what)
        if modules is None:
            return None
        
        public = [s for s in modules if s[0] != '_']
        private = [s for s in modules if s[0] == '_']
        is_case_insen = False
        return comp_prefix, public, private, is_case_insen
        
    def _complete_module_members(self, line, is_auto):
        """
        line - the stripped line from its beginning to the cursor.
        Return (comp_prefix, public, private, is_case_insen) 
        (string, list, list, bool).
        If shouldn't complete - return None.
        """
        # Check whether autocompletion is really appropriate
        if is_auto:
            return
        
        i = len(line)
        while i and line[i-1] in ID_CHARS:
            i -= 1
        comp_prefix = line[i:]
        
        m = re.match(r'from\s+([\w.]+)\s+import', line)
        if m is None:
            return
        comp_what = m.group(1)
        
        public_and_private = self.get_module_members(comp_what)
        if public_and_private is None:
            return
        public, private = public_and_private
        is_case_insen = False
        return comp_prefix, public, private, is_case_insen
        
    def _complete_filenames(self, text, index, hp, is_auto):
        """
        Return (comp_prefix, public, private, is_case_insen) 
        (string, list, list, bool).
        If shouldn't complete - return None.
        """
        # Check whether autocompletion is really appropriate
        if is_auto and text[index-1] not in '\\/':
            return
        
        str_start = hp.bracketing[hp.indexbracket][0] + 1
        # Analyze string a bit
        pos = str_start - 1
        str_char = text[pos]
        assert str_char in ('"', "'")
        if text[pos+1:pos+3] == str_char + str_char:
            # triple-quoted string - not for us
            return
        is_raw = pos > 0 and text[pos-1].lower() == 'r'
        if is_raw:
            pos -= 1
        is_unicode = pos > 0 and text[pos-1].lower() == 'u'
        if is_unicode:
            pos -= 1
        str_prefix = text[pos:str_start]

        # Do not open a completion list if after a single backslash in a
        # non-raw string
        if is_auto and text[index-1] == '\\' \
           and not is_raw and not self._is_backslash_char(text, index-1):
            return

        # Find completion start - last '/' or real '\\'
        sep_ind = max(text.rfind('/', 0, index), text.rfind('\\', 0, index))
        if sep_ind == -1 or sep_ind < str_start:
            # not found - prefix is all the string.
            comp_prefix_index = str_start
        elif text[sep_ind] == '\\' and not is_raw and not self._is_backslash_char(text, sep_ind):
            # Do not complete if the completion prefix contains a backslash.
            return
        else:
            comp_prefix_index = sep_ind+1

        comp_prefix = text[comp_prefix_index:index]
        
        add_quote = not (len(text) > index and text[index] == str_char)
        
        res = self.complete_filenames(
            str_prefix, text[str_start:comp_prefix_index], str_char, add_quote)
        if res is None:
            return
        public, private, is_case_insen = res
        
        return comp_prefix, public, private, is_case_insen
    
    def _on_complete(self):
        # Called when the user completed. This is relevant if he completed
        # a dir name, so that another completion window will be opened.
        self.show_completions(is_auto=True, complete=False)
        
    @staticmethod
    def _is_backslash_char(string, index):
        """
        Assuming that string[index] is a backslash, check whether it's a
        real backslash char or just an escape - if it has an odd number of
        preceding backslashes it's a real backslash
        """
        assert string[index] == '\\'
        count = 0
        while index-count > 0 and string[index-count-1] == '\\':
            count += 1
        return (count % 2) == 1



# This file was automatically generated from a file in the source DreamPie dir.
# DO NOT EDIT IT, as your changes will be gone when the file is created again.

__version__ = "1.2"
release_timestamp = 1348531200 # calendar.timegm((2012, 9, 25, 0, 0, 0))



# Copyright 2010 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

import sys
import os
from os.path import join, isdir, exists
import stat
import imp
import re
import time

TIMEOUT = 1 # Stop after 1 second

# Match any of the suffixes
suffix_re = re.compile(
    r'(?:%s)$' % '|'.join(re.escape(suffix[0]) for suffix in imp.get_suffixes()))

# A mapping from absolute names to (mtime, module_names) tuple.
cache = {}

def find_in_dir(dirname):
    """
    Yield all names of modules in the given dir.
    """
    if dirname == '':
        dirname = '.'
    try:
        basenames = os.listdir(dirname)
    except OSError:
        return
    for basename in basenames:
        m = suffix_re.search(basename)
        if m:
            yield basename[:m.start()]
        else:
            if '.' not in basename and isdir(join(dirname, basename)):
                init = join(dirname, basename, '__init__.py')
                if exists(init) or exists(init+'c'):
                    yield basename    

def find_in_dir_cached(dirname):
    if dirname not in cache:
        # If it is in cache, it's already absolute.
        dirname = os.path.abspath(dirname)
    try:
        st = os.stat(dirname)
    except OSError:
        return ()
    if not stat.S_ISDIR(st.st_mode):
        return ()
    try:
        mtime, modules = cache[dirname]
    except KeyError:
        mtime = 0
    if mtime != st.st_mtime:
        modules = list(find_in_dir(dirname))
        cache[dirname] = (st.st_mtime, modules)
    return modules

def find_package_path(package):
    """
    Get a package as a list, try to find its path (list of dirs) or return None.
    """
    for i in xrange(len(package), 0, -1):
        package_name = '.'.join(package[:i])
        if package_name in sys.modules:
            try:
                path = sys.modules[package_name].__path__
            except AttributeError:
                return None
            break
    else:
        i = 0
        path = sys.path
    
    for j in xrange(i, len(package)):
        name = package[j]
        for dir in path:
            newdir = join(dir, name)
            if isdir(newdir):
                path = [newdir]
                break
        else:
            return None
    
    return path

def find_modules(package):
    """
    Get a sequence of names (what you get from package_name.split('.')),
    or [] for a toplevel module.
    Return a list of module names.
    """
    start_time = time.time()
    r = set()
    path = find_package_path(package)
    if path:
        for dirname in path:
            r.update(find_in_dir_cached(dirname))
            if time.time() - start_time > TIMEOUT:
                break
    prefix = ''.join(s+'.' for s in package)
    for name in sys.modules:
        if name.startswith(prefix):
            mod = name[len(prefix):]
            if '.' not in mod:
                r.add(mod)
    r.discard('__init__')
    return sorted(r)



# Copyright 2012 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['bug_report', 'set_subp_pyexec']

import sys
import platform
import time
import webbrowser

import gtk
from gtk import glade

from .. import __version__
from .git import get_commit_details

subp_pyexec = None
subp_desc = None
def set_subp_info(pyexec, welcome):
    global subp_pyexec, subp_desc
    subp_pyexec = pyexec
    subp_desc = welcome.split('\n')[0]


_is_git = _latest_name = _latest_time = _cur_time = None
def set_update_info(is_git, latest_name, latest_time, cur_time):
    global _is_git, _latest_name, _latest_time, _cur_time
    _is_git = is_git
    _latest_name = latest_name
    _latest_time = latest_time
    _cur_time = cur_time

def get_prefilled(trace):
    commit_id, commit_time = get_commit_details()
    commit_date = time.strftime('%Y/%m/%d', time.localtime(commit_time))
    s = """\
What steps will reproduce the problem?
1.
2.
3.

What is the expected result?


What happens instead?


Please provide any additional information below. To submit a screenshot, you \
can go to imgur.com, upload the image, and paste the URL.





-------------------
Diagnostic information:

DreamPie version: {version}
git commit: {commit_id} from {commit_date}
platform: {platform}
architecture: {architecture}
python_version: {python_version}
python_implementation: {python_implementation}
executable: {executable}
subprocess executable: {subp_pyexec}
subprocess description: {subp_desc}

""".format(version=__version__,
           commit_id=commit_id,
           commit_date=commit_date,
           platform=platform.platform(),
           architecture=platform.architecture(),
           python_version=platform.python_version(),
           python_implementation=platform.python_implementation(),
           executable=sys.executable,
           subp_pyexec=subp_pyexec,
           subp_desc=subp_desc,
           )
    if trace:
        s += trace
    return s

def get_update_message():
    if _latest_time is None:
        return None
    if _latest_time <= _cur_time and _is_git:
        return None
    
    if _latest_time > _cur_time:
        if _is_git:
            msg = """\
Note: you are not using the latest git commit. Please run 'git pull' and see \
if the problem persists."""
        else:
            msg = """\
Note: you are using an out of date version of DreamPie. Please go to \
www.dreampie.org/download.html and download a new version.

If you can, please use a git repository. This will let you see if the bug was \
already fixed and let you check immediately if the committed fix actually \
works. You will also enjoy other improvements earlier."""
    else:
        msg = """\
Note: you are using the DreamPie released version. If you can, please clone \
the git repository from https://github.com/noamraph/dreampie.git and see if \
the problem persists. Even if it does, it will let you check immediately if \
the committed fix actually works. You will also enjoy other improvements \
earlier."""

    return '<span color="red">%s</span>\n' % msg

def bug_report(master_window, gladefile, trace):
    """
    Send the user to a bug report webpage, instructing him to paste a template
    with questions and diagnostics information.
    
    master_window: gtk.Window, master of the dialog.
    gladefile: glade filename, for getting the widgets.
    trace: a string with the formatted traceback, or None.
    """
    xml = glade.XML(gladefile, 'bug_report_dialog')
    d = xml.get_widget('bug_report_dialog')
    bug_report_textview = xml.get_widget('bug_report_textview')
    update_label = xml.get_widget('update_label')
    d.set_transient_for(master_window)
    d.set_default_response(gtk.RESPONSE_OK)
    
    prefilled = get_prefilled(trace)
    tb = bug_report_textview.get_buffer()
    tb.set_text(prefilled)
    update_msg = get_update_message()
    if update_msg:
        update_label.set_markup(update_msg)
        update_label.show()
    clipboard = gtk.Clipboard()
    clipboard.set_text(prefilled)
    
    r = d.run()
    d.destroy()
    
    if r == gtk.RESPONSE_OK:
        webbrowser.open('https://github.com/noamraph/dreampie/issues/new')



# Copyright 2009 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

"""
Tags for the textview and sourceview.
"""

import os
import tempfile

from gtk import gdk
import gtksourceview2

# DEFAULT is not really a tag, but it means the default text colors
DEFAULT = 'default'

# Tags for marking output
OUTPUT = 'output'
STDIN = 'stdin'; STDOUT = 'stdout'; STDERR = 'stderr'; EXCEPTION = 'exception'
RESULT_IND = 'result-ind'; RESULT = 'result'

# Tags for marking commands
PROMPT = 'prompt'; COMMAND = 'command'; COMMAND_DEFS='command-defs'
COMMAND_SEP = 'commandsep'

# Folding tags
FOLDED = 'folded'
FOLD_MESSAGE = 'fold-message'

# The MESSAGE tag
MESSAGE = 'message'

# Tags for syntax highlighting
KEYWORD = 'keyword'; BUILTIN = 'builtin'; STRING = 'string'
NUMBER = 'number'; COMMENT = 'comment'; BRACKET_MATCH = 'bracket-match'
BRACKET_1 = 'bracket-1'; BRACKET_2 = 'bracket-2'; BRACKET_3 = 'bracket-3'
ERROR = 'error'

# Constants to retrieve data from a theme. A theme is just a dict which maps
# tuples to strings, and is used like this: 
# theme[KEYWORD, FG, COLOR], theme[COMMENT, BG, ISSET]
FG = 'fg'; BG = 'bg'
COLOR = 'color'; ISSET = 'isset'

# Add this string to theme names to get the config section
THEME_POSTFIX = ' theme'

# Tags which affect appearence
tag_desc = [
    (DEFAULT, 'Default'),

    (KEYWORD, 'Keyword'),
    (BUILTIN, 'Builtin'),
    (STRING, 'String'),
    (NUMBER, 'Number'),
    (COMMENT, 'Comment'),
    (BRACKET_MATCH, 'Bracket Match'),
    (BRACKET_1, 'Bracket 1'),
    (BRACKET_2, 'Bracket 2'),
    (BRACKET_3, 'Bracket 3'),
    (ERROR, 'Error'),
    
    (STDIN, 'Standard Input'),
    (STDOUT, 'Standard Output'),
    (STDERR, 'Standard Error'),
    (RESULT, 'Result'),
    (RESULT_IND, 'Result Index'),
    (EXCEPTION, 'Exception'),
    (PROMPT, 'Prompt'),
    (MESSAGE, 'Messages'),
    (FOLD_MESSAGE, 'Folded Text'),
    ]

"""
Some documentation about what's going on in the text buffer
-----------------------------------------------------------

The text buffer has two purposes: to show the user what he expects, and to
store all the information needed about the history. So here I try to document
what's going on there.

Most tags can be considered to only affect the highlight color. Here only the
tags which affect the data model are described.

The COMMAND tag is applied to code segments. The last newline is also tagged.
After it comes a '\r' char marked with COMMAND_SEP, so it's invisible. It's
used to separate two code segments even if there was no output in
between. The prompt is marked with both the COMMAND and the PROMPT tag, so the
Copy Code Only action copies only text which is marked by COMMAND and not by
PROMPT (and adds a trailing newline.) lines inside def and class blocks are
also marked with COMMAND_DEFS, so they can be hidden if the user wants to.

STDIN is marked as STDIN, COMMAND and OUTPUT. It always ends with a \n. It is
marked with COMMAND so that it will be history-searched. It is marked with
OUTPUT so that it will be considered as output for folding purposes.

The MESSAGE tag tags a message which is displayed at the beginning of each
session (that is, subprocess). It's either the welcome message (Python version),
a "New Session" message or a "History Discarded" message. When you discard
previous sessions, the MESSAGE tag is used to understand what are the previous
sessions.

Text marked with OUTPUT was written by output.py. It includes stdout, stderr,
result and exception. This text is written at the *output mark*, which means
that if an output is produced after the code execution was finished (for
example, by another thread), it will appear before the prompt. Also, ANSI
escapes are removed. '\r' (carriage return) is handled (to a point - there's
no support for having a part of a line overwritten - after a line starts to be
overwritten, it's completely removed). When lines are too long (too many chars
without a '\n', they are broken to a certain maximum length by '\r' chars.
Original '\r' chars will never pass to the text buffer. These '\r' chars are
treated by the text buffer just like '\n' chars, but are not copied, because
they are not real chars - they were only inserted so that the text buffer won't
become unresponsive. If there are enough lines in an output section, it gets
automatically collapsed. There's always a '\n' after the output section which
is marked as OUTPUT (unless nothing was written, which is not really an
output section.)

Folded output/code sections look like this:

First few characters of the output/code section, which are truncated somewhe
(The rest of the section, tagged with FOLDED so invisible)
[About n more lines. Double-click to unfold]

* From some point (which will be a '\n' if the first line isn't too long), the
  output is marked with FOLDED, so it is invisible.
* The '\n' which always comes at the end of the output is also invisible. It
  doesn't make a lot of sense, because that way no newline is visible, but it
  turns out that if it's visible, you get an extra empty line.
* Afterwards comes the FOLD_MESSAGE, which includes a '\n' at the end so that
  the background color is behind the entire line.
"""

def get_theme_names(config):
    for section in config.sections():
        if section.endswith(THEME_POSTFIX):
            if config.get_bool('is-active', section):
                yield section[:-len(THEME_POSTFIX)]

def get_theme(config, theme_name):
    """
    Get a theme description (a dict of tuples, see above) from a config object.
    """
    section = theme_name + THEME_POSTFIX
    if not config.get_bool('is-active', section):
        raise ValueError("Theme %s is not active" % theme_name)
    theme = {}
    for tag, _desc in tag_desc:
        theme[tag, FG, COLOR] = config.get('%s-fg' % tag, section)
        theme[tag, BG, COLOR] = config.get('%s-bg' % tag, section)
        if tag != DEFAULT:
            theme[tag, FG, ISSET] = config.get_bool('%s-fg-set' % tag, section)
            theme[tag, BG, ISSET] = config.get_bool('%s-bg-set' % tag, section)
    return theme

def set_theme(config, theme_name, theme):
    """
    Write a theme description to a config object.
    """
    section = theme_name + THEME_POSTFIX
    if not config.has_section(section):
        config.add_section(section)
    config.set_bool('is-active', True, section)
    for tag, _desc in tag_desc:
        config.set('%s-fg' % tag, theme[tag, FG, COLOR], section)
        config.set('%s-bg' % tag, theme[tag, BG, COLOR], section)
    for tag, _desc in tag_desc:
        if tag != DEFAULT:
            config.set_bool('%s-fg-set' % tag, theme[tag, FG, ISSET], section)
            config.set_bool('%s-bg-set' % tag, theme[tag, BG, ISSET], section)

def remove_themes(config):
    """
    Remove all themes.
    """
    for name in get_theme_names(config):
        # We replace the section with a section with 'is-active = False', so
        # that if the section is updated from default configuration values
        # it will not reappear.
        section = name + THEME_POSTFIX
        config.remove_section(section)
        config.add_section(section)
        config.set_bool('is-active', False, section)

def get_actual_color(theme, tag, fg_or_bg):
    """
    Get the actual color that will be displayed - taking ISSET into account.
    """
    if tag == DEFAULT or theme[tag, fg_or_bg, ISSET]:
        return theme[tag, fg_or_bg, COLOR]
    else:
        return theme[DEFAULT, fg_or_bg, COLOR]

def add_tags(textbuffer):
    """
    Add the needed tags to a textbuffer
    """
    for tag, _desc in tag_desc:
        if tag != DEFAULT:
            textbuffer.create_tag(tag)
    textbuffer.create_tag(OUTPUT)
    textbuffer.create_tag(COMMAND)
    textbuffer.create_tag(COMMAND_DEFS)
    tag = textbuffer.create_tag(COMMAND_SEP)
    tag.props.invisible = True
    tag = textbuffer.create_tag(FOLDED)
    tag.props.invisible = True

def apply_theme_text(textview, textbuffer, theme):
    """
    Apply the theme to the textbuffer. add_tags should have been called
    previously.
    """
    for tag, _desc in tag_desc:
        if tag == DEFAULT:
            textview.modify_base(0, gdk.color_parse(theme[tag, BG, COLOR]))
            textview.modify_text(0, gdk.color_parse(theme[tag, FG, COLOR]))
        else:
            tt = textbuffer.get_tag_table().lookup(tag)
            tt.props.foreground = theme[tag, FG, COLOR]
            tt.props.foreground_set = theme[tag, FG, ISSET]
            tt.props.background = theme[tag, BG, COLOR]
            tt.props.background_set = theme[tag, BG, ISSET]
            tt.props.paragraph_background = theme[tag, BG, COLOR]
            tt.props.paragraph_background_set = theme[tag, BG, ISSET]

def _make_style_scheme(spec):
    # Quite stupidly, there's no way to create a SourceStyleScheme without
    # reading a file from a search path. So this function creates a file in
    # a directory, to get you your style scheme.
    #
    # spec should be a dict of dicts, mapping style names to (attribute, value)
    # pairs. Color values will be converted using gdk.color_parse().
    # Boolean values will be handled correctly.
    dir = tempfile.mkdtemp()
    filename = os.path.join(dir, 'scheme.xml')
    f = open(filename, 'w')
    f.write('<?xml version="1.0" encoding="UTF-8"?>\n')
    f.write('<style-scheme id="scheme" _name="Scheme" version="1.0">\n')
    for name, attributes in spec.iteritems():
        f.write('<style name="%s" ' % name)
        for attname, attvalue in attributes.iteritems():
            if attname in ('foreground', 'background'):
                attvalue = gdk.color_parse(attvalue).to_string()
            elif attname in ('italic', 'bold', 'underline', 'strikethrough',
                             'foreground-set', 'background-set'):
                attvalue = 'true' if attvalue else 'false'
            f.write('%s="%s" ' % (attname, attvalue))
        f.write('/>\n')
    f.write('</style-scheme>\n')
    f.close()
    
    ssm = gtksourceview2.StyleSchemeManager()
    ssm.set_search_path([dir])
    scheme = ssm.get_scheme('scheme')

    os.remove(filename)
    os.rmdir(dir)

    return scheme

def _get_style_scheme_spec(theme):
    mapping = {
        'text': DEFAULT,
        
        'def:keyword': KEYWORD,
        'def:preprocessor': KEYWORD,

        'def:builtin': BUILTIN,
        'def:special-constant': BUILTIN,
        'def:type': BUILTIN,

        'def:string': STRING,
        'def:number': NUMBER,
        'def:comment': COMMENT,

        'bracket-match': BRACKET_MATCH,
        'python:bracket-1': BRACKET_1,
        'python:bracket-2': BRACKET_2,
        'python:bracket-3': BRACKET_3,
        'def:error': ERROR,
        }

    res = {}
    for key, value in mapping.iteritems():
        res[key] = dict(foreground=get_actual_color(theme, value, FG),
                        background=get_actual_color(theme, value, BG))
    return res

def apply_theme_source(sourcebuffer, theme):
    sourcebuffer.set_style_scheme(
        _make_style_scheme(_get_style_scheme_spec(theme)))





# Copyright 2012 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

all = ['get_commit_details']

from os.path import abspath, join, dirname, isdir

def get_commit_details():
    """
    If there's a '.git' directory besides 'dreampielib', return the current
    commit (HEAD) id and commit time.
    Otherwise, return None, None.
    """
    git_dir = join(dirname(dirname(dirname(abspath(__file__)))), '.git')
    if not isdir(git_dir):
        return None, None
    try:
        from dulwich.repo import Repo
    except ImportError:
        return None, None
    
    repo = Repo(git_dir)
    commit_id = repo.refs['HEAD']
    commit = repo.commit(commit_id)
    return commit_id, commit.commit_time



# Copyright 2010 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

all = ['Folding']

from .tags import OUTPUT, COMMAND, FOLDED, FOLD_MESSAGE
from .common import beep, get_text

# Maybe someday we'll want translations...
_ = lambda s: s

class Folding(object):
    """
    Support folding and unfolding of output and code sections.
    """
    def __init__(self, textbuffer, LINE_LEN):
        self.textbuffer = tb = textbuffer
        self.LINE_LEN = LINE_LEN
        
        # Mark the bottom-most section which was unfolded, so as not to
        # auto-fold it.
        self.last_unfolded_mark = tb.create_mark(
            'last-folded', tb.get_start_iter(), left_gravity=True)
        
        tt = self.textbuffer.get_tag_table()
        self.fold_message_tag = tt.lookup(FOLD_MESSAGE)
        self.output_tag = tt.lookup(OUTPUT)
        self.command_tag = tt.lookup(COMMAND)
        self.tags = {OUTPUT: self.output_tag, COMMAND: self.command_tag}
    
    def get_section_status(self, it):
        """
        Get an iterator of the sourcebuffer. Return a tuple:
        (typ, is_folded, start_it)
        typ: one of tags.OUTPUT, tags.COMMAND
        is_folded: boolean (is folded), or None if not folded but too short
                   to fold (1 line or less).
        start_it: An iterator pointing to the beginning of the section.
        
        If it isn't in an OUTPUT or COMMAND section, return None.
        """
        it = it.copy()
        # The iterator is in an OUTPUT section if it's either tagged with
        # OUTPUT or if it's inside a FOLD_MESSAGE which goes right after
        # the OUTPUT tagged text. The same goes for COMMAND - note that STDIN
        # is marked with both COMMAND and OUTPUT and is considered output, so
        # we check OUTPUT first.
        # A section is folded iff it's followed by a FOLD_MESSAGE.
        if it.has_tag(self.fold_message_tag):
            if not it.begins_tag(self.fold_message_tag):
                it.backward_to_tag_toggle(self.fold_message_tag)
            if it.ends_tag(self.output_tag):
                typ = OUTPUT
            elif it.ends_tag(self.command_tag):
                typ = COMMAND
            else:
                assert False, "FOLD_MESSAGE doesn't follow OUTPUT/COMMAND"
            it.backward_to_tag_toggle(self.tags[typ])
            return (typ, True, it)
        else:
            if it.has_tag(self.output_tag) or it.ends_tag(self.output_tag):
                typ = OUTPUT
                tag = self.output_tag
            elif it.has_tag(self.command_tag) or it.ends_tag(self.command_tag):
                typ = COMMAND
                tag = self.command_tag
            else:
                return None
            if not it.ends_tag(tag):
                it.forward_to_tag_toggle(tag)
            end_it = it.copy()
            is_folded = end_it.has_tag(self.fold_message_tag)
            it.backward_to_tag_toggle(tag)
            if not is_folded:
                n_lines = self._count_lines(it, end_it)
                if n_lines <= 1:
                    is_folded = None
            return (typ, is_folded, it)
    
    def _count_lines(self, start_it, end_it):
        return max(end_it.get_line()-start_it.get_line(),
                   (end_it.get_offset()-start_it.get_offset())//self.LINE_LEN)
    
    def fold(self, typ, start_it):
        """
        Get an iterator pointing to the beginning of an unfolded OUTPUT/COMMAND
        section. Fold it.
        """
        tb = self.textbuffer
        
        # Move end_it to the end of the section
        end_it = start_it.copy()
        end_it.forward_to_tag_toggle(self.tags[typ])
        n_lines = self._count_lines(start_it, end_it)
        
        # Move 'it' to the end of the first line (this is where we start hiding)
        it = start_it.copy()
        it.forward_chars(self.LINE_LEN)
        first_line = get_text(tb, start_it, it)
        newline_pos = first_line.find('\n')
        if newline_pos != -1:
            it.backward_chars(len(first_line)-newline_pos)
        
        # Hide
        tb.apply_tag_by_name(FOLDED, it, end_it)
        
        # Add message
        tb.insert_with_tags_by_name(
            end_it,
            _("[About %d more lines. Double-click to unfold]\n") % (n_lines-1),
            FOLD_MESSAGE)
    
    def unfold(self, typ, start_it):
        """
        Get an iterator pointing to the beginning of an unfolded OUTPUT/COMMAND
        section. Unfold it.
        """
        tb = self.textbuffer
        
        last_unfolded_it = tb.get_iter_at_mark(self.last_unfolded_mark)
        if start_it.compare(last_unfolded_it) > 0:
            tb.move_mark(self.last_unfolded_mark, start_it)
    
        it = start_it.copy()
        it.forward_to_tag_toggle(self.tags[typ])
        tb.remove_tag_by_name(FOLDED, start_it, it)
        
        it2 = it.copy()
        it2.forward_to_tag_toggle(self.fold_message_tag)
        assert it2.ends_tag(self.fold_message_tag)
        tb.delete(it, it2)
        
    def autofold(self, it, numlines):
        """
        Get an iterator to a recently-written output section.
        If it is folded, update the fold message and hide what was written.
        It it isn't folded, then if the number of lines exceeds numlines and
        the section wasn't manually unfolded, fold it.
        """
        tb = self.textbuffer
        
        typ, is_folded, start_it = self.get_section_status(it)
        if is_folded:
            # Just unfold and fold. We create a mark because start_iter is
            # invalidated
            start_it_mark = tb.create_mark(None, start_it, left_gravity=True)
            self.unfold(typ, start_it)
            start_it = tb.get_iter_at_mark(start_it_mark)
            tb.delete_mark(start_it_mark)
            self.fold(typ, start_it)
        else:
            last_unfolded_it = tb.get_iter_at_mark(self.last_unfolded_mark)
            if not start_it.equal(last_unfolded_it):
                end_it = start_it.copy()
                end_it.forward_to_tag_toggle(self.tags[typ])
                n_lines = self._count_lines(start_it, end_it)
                if n_lines >= numlines:
                    self.fold(typ, start_it)
    
    def get_tag(self, typ):
        """Return the gtk.TextTag for a specific typ string."""
        return self.tags[typ]
    
    def fold_last(self):
        """
        Fold last unfolded output section.
        """
        tb = self.textbuffer
        it = tb.get_end_iter()

        while True:
            r = it.backward_to_tag_toggle(self.output_tag)
            if not r:
                # Didn't find something to fold
                beep()
                break
            if it.begins_tag(self.output_tag):
                typ, is_folded, start_it = self.get_section_status(it)
                if is_folded is not None and not is_folded:
                    self.fold(typ, start_it)
                    break
                

    def unfold_last(self):
        """
        Unfold last folded output section.
        """
        tb = self.textbuffer
        it = tb.get_end_iter()
        
        while True:
            r = it.backward_to_tag_toggle(self.output_tag)
            if not r:
                # Didn't find something to fold
                beep()
                return
            if not it.begins_tag(self.output_tag):
                continue
            typ, is_folded, start_it = self.get_section_status(it)
            if is_folded:
                self.unfold(typ, start_it)
                return



# Copyright 2009 Noam Yorav-Raphael
#
# This file is part of DreamPie.
# 
# DreamPie is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# DreamPie is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with DreamPie.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['SubprocessHandler']

import sys
import os
import time
if sys.platform != 'win32':
    import signal
else:
    import ctypes
from .subprocess_interact import Popen, PIPE
from select import select
import socket
import random
from logging import debug

import gobject

from ..common.objectstream import send_object, recv_object

_ = lambda s: s

START_TIMEOUT = 30 # seconds

class StartError(IOError):
    """Error when starting subprocess"""
    pass

class StartTerminatedError(StartError):
    """Start subprocess failed because process terminated."""
    def __init__(self, rc, output):
        self.rc = rc
        self.output = output
    def __str__(self):
        r = _("Subprocess terminated with return code %d.") % self.rc
        if self.output:
            r += _("\nSubprocess wrote:\n%s") % self.output
        return r

class StartTimeoutError(StartError):
    """Start subprocess failed because timeout elapsed."""
    def __init__(self, timeout, output):
        self.timeout = timeout
        self.output = output
    def __str__(self):
        r = _("Subprocess didn't call back in %s seconds. "
              "This may be caused by a firewall program blocking the "
              "communication between the main process and the subproces."
              ) % self.timeout
        if self.output:
            r += _("\nSubprocess wrote:\n%s") % self.output
        return r

class SubprocessHandler(object):
    """
    Manage interaction with the subprocess.
    The communication, besides stdout, stderr and stdin, goes like this:
    You can call a function, and get a return value.
    (This sends over the tuple with the function name and parameters,
    and waits for the next object, which is the return value.)
    You can also get objects asyncronically.
    (This happens when not waiting for a function's return value.)
    """

    def __init__(self, pyexec, data_dir,
                 on_stdout_recv, on_stderr_recv, on_object_recv,
                 on_subp_terminated):
        self._pyexec = pyexec
        self._data_dir = data_dir
        self._on_stdout_recv = on_stdout_recv
        self._on_stderr_recv = on_stderr_recv
        self._on_object_recv = on_object_recv
        self._on_subp_terminated = on_subp_terminated
        
        self._sock = None
        # self._popen is None when there's no subprocess
        self._popen = None
        self._last_kill_time = 0
        
        # I know that polling isn't the best way, but on Windows you have
        # no choice, and it allows us to do it all by ourselves, not use
        # gobject's functionality.
        gobject.timeout_add(10, self._manage_subp)
        
    def start(self):
        if self._popen is not None:
            raise ValueError("Subprocess is already living")
        # Find a socket to listen to
        ports = range(10000, 10100)
        random.shuffle(ports)
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        for port in ports:
            #debug("Trying to listen on port %d..." % port)
            try:
                s.bind(('localhost', port))
            except socket.error:
                #debug("Failed.")
                pass
            else:
                #debug("Ok.")
                break
        else:
            raise IOError("Couldn't find a port to bind to")
        # Now the socket is bound to port.

        #debug("Spawning subprocess")
        env = os.environ.copy()
        env['PYTHONUNBUFFERED'] = '1'
        env['PYTHONIOENCODING'] = 'UTF-8'
        script = os.path.join(self._data_dir, 'subp_main.py')
        # The -S switch causes the subprocess to not automatically import
        # site.py. This is done so that the subprocess will be able to call
        # sys.setdefaultencoding('UTF-8') before importing site, and this is
        # needed because Python 2.5 ignores the PYTHONIOENCODING variable.
        # Hopefully it won't cause problems.
        popen = Popen([self._pyexec, '-S', script, str(port)],
                       stdin=PIPE, stdout=PIPE, stderr=PIPE,
                       env=env)
        #debug("Waiting for the subprocess to connect")
        s.listen(1)
        # We wait for the client to connect, but we also poll stdout and stderr,
        # and if it writes something then we report an error.
        s.settimeout(0.1)
        start_time = time.time()
        while True:
            try:
                self._sock, _addr = s.accept()
            except socket.timeout:
                pass
            else:
                break
            
            rc = popen.poll()
            if rc is not None:
                out = (popen.recv() or '') + (popen.recv_err() or '')
                raise StartTerminatedError(rc, out)
            
            if time.time() - start_time > START_TIMEOUT:
                out = (popen.recv() or '') + (popen.recv_err() or '')
                raise StartTimeoutError(START_TIMEOUT, out)
        self._sock.setblocking(True)
            
        #debug("Connected to addr %r." % (addr,))
        s.close()
        self._popen = popen

    def _manage_subp(self):
        popen = self._popen
        if popen is None:
            # Just continue looping - there's no subprocess.
            return True

        # Check if exited
        rc = popen.poll()
        if rc is not None:
            if time.time() - self._last_kill_time > 10:
                debug("Process terminated unexpectedly with rc %r" % rc)
            self._sock.close()
            self._sock = None
            self._popen = None
            self._on_subp_terminated()
            return True

        # Read from stdout
        r = popen.recv()
        if r:
            self._on_stdout_recv(r.decode('utf8', 'replace'))

        # Read from stderr
        r = popen.recv_err()
        if r:
            self._on_stderr_recv(r.decode('utf8', 'replace'))
        
        # Read from socket
        if self.wait_for_object(0):
            try:
                obj = recv_object(self._sock)
            except IOError:
                # Could happen when subprocess exits. See bug #525358.
                # We give the subprocess a second. If it shuts down, we ignore
                # the exception, since on the next round we will handle that.
                # Otherwise, the exception remains unexplained so we re-raise.
                time.sleep(1)
                if popen.poll() is None:
                    raise
            else:
                self._on_object_recv(obj)

        return True

    def send_object(self, obj):
        """Send an object to the subprocess"""
        if self._popen is None:
            raise ValueError("Subprocess not living")
        send_object(self._sock, obj)

    def wait_for_object(self, timeout_s):
        """
        Wait for timeout_s seconds or until the socket is ready for reading.
        Return True if an object was received, and False if the timeout expired.
        """
        return len(select([self._sock], [], [], timeout_s)[0]) > 0
    
    def recv_object(self):
        """Wait for an object from the subprocess and return it"""
        if self._popen is None:
            raise ValueError("Subprocess not living")
        return recv_object(self._sock)

    def write(self, data):
        """Write data to stdin"""
        if self._popen is None:
            raise ValueError("Subprocess not living")
        self._popen.stdin.write(data.encode('utf8'))

    def kill(self):
        """Kill the subprocess.
        If the event loop continues, will start another one."""
        if self._popen is None:
            raise ValueError("Subprocess not living")
        if sys.platform != 'win32':
            # Send SIGTERM, and if the process didn't terminate within 1 second,
            # send SIGKILL.
            os.kill(self._popen.pid, signal.SIGTERM)
            killtime = time.time()
            while True:
                rc = self._popen.poll()
                if rc is not None:
                    break
                if time.time() - killtime > 1:
                    os.kill(self._popen.pid, signal.SIGKILL)
                    break
                time.sleep(0.1)
        else:
            kernel32 = ctypes.windll.kernel32
            PROCESS_TERMINATE = 1
            handle = kernel32.OpenProcess(PROCESS_TERMINATE, False,
                                          self._popen.pid)
            kernel32.TerminateProcess(handle, -1)
            kernel32.CloseHandle(handle)
        self._last_kill_time = time.time()

    def interrupt(self):
        if self._popen is None:
            raise ValueError("Subprocess not living")
        if sys.platform != 'win32':
            os.kill(self._popen.pid, signal.SIGINT)
        else:
            kernel32 = ctypes.windll.kernel32
            CTRL_C_EVENT = 0
            try:
                kernel32.GenerateConsoleCtrlEvent(CTRL_C_EVENT, 0)
                time.sleep(10)
            except KeyboardInterrupt:
                # This also sends us a KeyboardInterrupt. It should
                # happen in time.sleep.
                pass

